from unittest.mock import MagicMock, call
import re
import math

import pytest

from qcodes.instrument_drivers.Keysight.keysightb1500 import constants
from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1517A import \
    B1517A
from qcodes.instrument_drivers.Keysight.keysightb1500.constants import \
    VOutputRange, CompliancePolarityMode, IOutputRange, IMeasRange, \
    VMeasRange, MM

# pylint: disable=redefined-outer-name


@pytest.fixture
def mainframe():
    yield MagicMock()


@pytest.fixture
def smu(mainframe):
    slot_nr = 1
    smu = B1517A(parent=mainframe, name='B1517A', slot_nr=slot_nr)
    yield smu


def test_snapshot():
    from qcodes.instrument.base import InstrumentBase
    # We need to use `InstrumentBase` (not a bare mock) in order for
    # `snapshot` methods call resolution to work out
    mainframe = InstrumentBase(name='mainframe')
    mainframe.write = MagicMock()
    slot_nr = 1
    smu = B1517A(parent=mainframe, name='B1517A', slot_nr=slot_nr)

    smu.use_high_speed_adc()
    smu.source_config(output_range=VOutputRange.AUTO)
    smu.i_measure_range_config(i_measure_range=IMeasRange.AUTO)
    smu.v_measure_range_config(v_measure_range=VMeasRange.AUTO)
    smu.timing_parameters(0.0, 0.123, 321)

    s = smu.snapshot()

    assert '_source_config' in s
    assert 'output_range' in s['_source_config']
    assert isinstance(s['_source_config']['output_range'], VOutputRange)
    assert '_measure_config' in s
    assert 'v_measure_range' in s['_measure_config']
    assert 'i_measure_range' in s['_measure_config']
    assert isinstance(s['_measure_config']['v_measure_range'], VMeasRange)
    assert isinstance(s['_measure_config']['i_measure_range'], IMeasRange)
    assert '_timing_parameters' in s
    assert 'number' in s['_timing_parameters']
    assert isinstance(s['_timing_parameters']['number'], int)


@pytest.mark.filterwarnings("ignore:The function <measure_config>")
def test_measure_config(smu):
    smu.measure_config(VMeasRange.AUTO)
    s = smu.snapshot()

    assert s['_measure_config']['v_measure_range'] == 0
    assert s['_measure_config']['i_measure_range'] == 0

    smu.measure_config(VMeasRange.FIX_0V5)
    s = smu.snapshot()

    assert isinstance(s['_measure_config']['v_measure_range'], VMeasRange)
    assert s['_measure_config']['v_measure_range'] == -5

    smu.measure_config(IMeasRange.FIX_1nA)
    s = smu.snapshot()

    assert isinstance(s['_measure_config']['i_measure_range'], IMeasRange)
    assert s['_measure_config']['i_measure_range'] == -11

    smu.measure_config(IMeasRange.AUTO)
    s = smu.snapshot()

    assert s['_measure_config']['v_measure_range'] == 0
    assert s['_measure_config']['i_measure_range'] == 0


def test_v_measure_range_config_raises_type_error(smu):
    msg = re.escape("Expected valid voltage measurement range, got 42.")

    with pytest.raises(TypeError, match=msg):
        smu.v_measure_range_config(v_measure_range=42)


def test_v_measure_range_config_raises_invalid_range_error(smu):
    msg = re.escape("15000 voltage measurement range")
    with pytest.raises(RuntimeError, match=msg):
        smu.v_measure_range_config(VMeasRange.MIN_1500V)


def test_v_measure_range_config_sets_range_correctly(smu):
    smu.v_measure_range_config(v_measure_range=VMeasRange.MIN_0V5)
    s = smu.snapshot()

    assert isinstance(s['_measure_config']['v_measure_range'], VMeasRange)
    assert s['_measure_config']['v_measure_range'] == 5


def test_getting_voltage_after_calling_v_measure_range_config(smu):
    mainframe = smu.parent
    mainframe.ask.return_value = "NAV-000.002E-01\r"

    smu.v_measure_range_config(VMeasRange.FIX_2V)

    assert smu.voltage.measurement_status is None
    assert pytest.approx(-0.2e-3) == smu.voltage()
    assert smu.voltage.measurement_status == constants.MeasurementStatus.N

    s = smu.voltage.snapshot()
    assert s


def test_i_measure_range_config_raises_type_error(smu):
    msg = re.escape("Expected valid current measurement range, got 99.")

    with pytest.raises(TypeError, match=msg):
        smu.i_measure_range_config(i_measure_range=99)


def test_i_measure_range_config_raises_invalid_range_error(smu):
    msg = re.escape("-23 current measurement range")
    with pytest.raises(RuntimeError, match=msg):
        smu.i_measure_range_config(IMeasRange.FIX_40A)


def test_i_measure_range_config_sets_range_correctly(smu):
    smu.i_measure_range_config(i_measure_range=IMeasRange.MIN_1nA)
    s = smu.snapshot()

    assert isinstance(s['_measure_config']['i_measure_range'], IMeasRange)
    assert s['_measure_config']['i_measure_range'] == 11


def test_getting_current_after_calling_i_measure_range_config(smu):
    mainframe = smu.parent
    mainframe.ask.return_value = "NAI+000.005E-06\r"

    smu.i_measure_range_config(IMeasRange.MIN_100mA)

    assert smu.current.measurement_status is None
    assert pytest.approx(0.005e-6) == smu.current()
    assert smu.current.measurement_status == constants.MeasurementStatus.N

    s = smu.current.snapshot()
    assert s


def test_force_invalid_voltage_output_range(smu):
    msg = re.escape("Invalid Source Voltage Output Range")
    with pytest.raises(RuntimeError, match=msg):
        smu.source_config(VOutputRange.MIN_1500V)


def test_force_invalid_current_output_range(smu):
    msg = re.escape("Invalid Source Current Output Range")
    with pytest.raises(RuntimeError, match=msg):
        smu.source_config(IOutputRange.MIN_20A)


def test_force_voltage_with_autorange(smu):
    mainframe = smu.parent

    smu.source_config(output_range=VOutputRange.AUTO)
    smu.voltage(10)

    mainframe.write.assert_called_once_with('DV 1,0,10')


def test_force_voltage_autorange_and_compliance(smu):
    mainframe = smu.parent

    smu.source_config(output_range=VOutputRange.AUTO,
                      compliance=1e-6,
                      compl_polarity=CompliancePolarityMode.AUTO,
                      min_compliance_range=IOutputRange.MIN_10uA)
    smu.voltage(20)

    mainframe.write.assert_called_once_with('DV 1,0,20,1e-06,0,15')


def test_new_source_config_should_invalidate_old_source_config(smu):
    mainframe = smu.parent

    smu.source_config(output_range=VOutputRange.AUTO,
                      compliance=1e-6,
                      compl_polarity=CompliancePolarityMode.AUTO,
                      min_compliance_range=IOutputRange.MIN_10uA)

    smu.source_config(output_range=VOutputRange.AUTO)
    smu.voltage(20)

    mainframe.write.assert_called_once_with('DV 1,0,20')


def test_unconfigured_source_defaults_to_autorange_v(smu):
    mainframe = smu.parent

    smu.voltage(10)

    mainframe.write.assert_called_once_with('DV 1,0,10')


def test_unconfigured_source_defaults_to_autorange_i(smu):
    mainframe = smu.parent

    smu.current(0.2)

    mainframe.write.assert_called_once_with('DI 1,0,0.2')


def test_force_current_with_autorange(smu):
    mainframe = smu.parent

    smu.source_config(output_range=IOutputRange.AUTO)
    smu.current(0.1)

    mainframe.write.assert_called_once_with('DI 1,0,0.1')


def test_raise_warning_output_range_mismatches_output_command(smu):
    smu.source_config(output_range=VOutputRange.AUTO)
    msg = re.escape("Asking to force current, but source_config contains a "
                    "voltage output range")
    with pytest.raises(TypeError, match=msg):
        smu.current(0.1)

    smu.source_config(output_range=IOutputRange.AUTO)
    msg = re.escape("Asking to force voltage, but source_config contains a "
                    "current output range")
    with pytest.raises(TypeError, match=msg):
        smu.voltage(0.1)


def test_measure_current(smu):
    mainframe = smu.parent
    mainframe.ask.return_value = "NAI+000.005E-06\r"

    assert smu.current.measurement_status is None

    assert pytest.approx(0.005e-6) == smu.current()
    assert smu.current.measurement_status == constants.MeasurementStatus.N


def test_measure_voltage(smu):
    mainframe = smu.parent
    mainframe.ask.return_value = "NAV+000.123E-06\r"

    assert smu.voltage.measurement_status is None

    assert pytest.approx(0.123e-6) == smu.voltage()
    assert smu.voltage.measurement_status == constants.MeasurementStatus.N

    s = smu.voltage.snapshot()
    assert s


def test_measure_current_shows_compliance_hit(smu):
    mainframe = smu.parent
    mainframe.ask.return_value = "CAI+000.123E-06\r"

    assert smu.current.measurement_status is None

    assert pytest.approx(0.123e-6) == smu.current()
    assert smu.current.measurement_status == constants.MeasurementStatus.C


def test_measured_voltage_with_V_status_returns_nan(smu):
    mainframe = smu.parent
    mainframe.ask.return_value = "VAV+199.999E+99\r"

    assert smu.voltage.measurement_status is None

    assert math.isnan(smu.voltage())
    assert smu.voltage.measurement_status == constants.MeasurementStatus.V


def test_some_voltage_sourcing_and_current_measurement(smu):
    mainframe = smu.parent

    smu.source_config(output_range=VOutputRange.MIN_0V5, compliance=1e-9)
    smu.i_measure_range_config(IMeasRange.FIX_100nA)

    mainframe.ask.return_value = "NAI+000.005E-09\r"

    smu.voltage(6)

    mainframe.write.assert_called_once_with('DV 1,5,6,1e-09')

    assert pytest.approx(0.005e-9) == smu.current()

    assert smu.voltage.measurement_status is None
    assert smu.current.measurement_status == constants.MeasurementStatus.N


def test_use_high_resolution_adc(smu):
    mainframe = smu.parent

    smu.use_high_resolution_adc()

    mainframe.write.assert_called_once_with('AAD 1,1')


def test_use_high_speed_adc(smu):
    mainframe = smu.parent

    smu.use_high_speed_adc()

    mainframe.write.assert_called_once_with('AAD 1,0')


def test_measurement_mode_at_init(smu):
    mode_at_init = smu.measurement_mode()
    assert mode_at_init == MM.Mode.SPOT


def test_measurement_mode_to_enum_value(smu):
    mainframe = smu.parent

    smu.measurement_mode(MM.Mode.SAMPLING)
    mainframe.write.assert_called_once_with('MM 10,1')

    new_mode = smu.measurement_mode()
    assert new_mode == MM.Mode.SAMPLING


def test_measurement_mode_to_int_value(smu):
    mainframe = smu.parent

    smu.measurement_mode(10)
    mainframe.write.assert_called_once_with('MM 10,1')

    new_mode = smu.measurement_mode()
    assert new_mode == MM.Mode.SAMPLING


def test_setting_timing_parameters(smu):
    mainframe = smu.parent

    smu.timing_parameters(0.0, 0.42, 32)
    mainframe.write.assert_called_once_with('MT 0.0,0.42,32')

    mainframe.reset_mock()

    smu.timing_parameters(0.0, 0.42, 32, 0.02)
    mainframe.write.assert_called_once_with('MT 0.0,0.42,32,0.02')


def test_set_average_samples_for_high_speed_adc(smu):
    mainframe = smu.parent

    smu.set_average_samples_for_high_speed_adc(131, 2)
    mainframe.write.assert_called_once_with('AV 131,2')

    mainframe.reset_mock()

    smu.set_average_samples_for_high_speed_adc(132)
    mainframe.write.assert_called_once_with('AV 132,0')



def test_measurement_operation_mode(smu):
    mainframe = smu.parent

    smu.measurement_operation_mode(constants.CMM.Mode.COMPLIANCE_SIDE)
    mainframe.write.assert_called_once_with('CMM 1,0')

    mainframe.reset_mock()

    mainframe.ask.return_value = 'CMM 1,0'
    cmm_mode = smu.measurement_operation_mode()
    assert cmm_mode == [(constants.ChNr.SLOT_01_CH1,
                         constants.CMM.Mode.COMPLIANCE_SIDE)]


def test_current_measurement_range(smu):
    mainframe = smu.parent

    smu.current_measurement_range(constants.IMeasRange.FIX_1A)
    mainframe.write.assert_called_once_with('RI 1,-20')

    mainframe.reset_mock()

    mainframe.ask.return_value = 'RI 1,-20'
    cmm_mode = smu.current_measurement_range()
    assert cmm_mode == [(constants.ChNr.SLOT_01_CH1,
                         constants.IMeasRange.FIX_1A)]


def test_get_sweep_mode_range_start_end_steps(smu):
    mainframe = smu.parent
    mainframe.ask.return_value = 'WV1,1,50,+3.0E+00,-3.0E+00,201'

    sweep_mode = smu.iv_sweep.sweep_mode()
    assert constants.SweepMode(1) == sweep_mode

    mainframe.reset_mock()

    sweep_range = smu.iv_sweep.sweep_range()
    assert constants.VOutputRange(50) == sweep_range

    sweep_start = smu.iv_sweep.sweep_start()
    assert 3.0 == sweep_start

    sweep_start = smu.iv_sweep.sweep_end()
    assert -3.0 == sweep_start

    sweep_start = smu.iv_sweep.sweep_steps()
    assert 201 == sweep_start

    current_compliance = smu.iv_sweep.current_compliance()
    assert current_compliance is None

def test_iv_sweep_delay(smu):
    mainframe = smu.root_instrument

    smu.iv_sweep.hold_time(43.12)
    smu.iv_sweep.delay(34.01)
    smu.iv_sweep.step_delay(0.01)
    smu.iv_sweep.trigger_delay(0.1)
    smu.iv_sweep.measure_delay(15.4)

    mainframe.write.assert_has_calls([call("WT 43.12,0.0,0.0,0.0,0.0"),
                                      call("WT 43.12,34.01,0.0,0.0,0.0"),
                                      call("WT 43.12,34.01,0.01,0.0,0.0"),
                                      call("WT 43.12,34.01,0.01,0.1,0.0"),
                                      call("WT 43.12,34.01,0.01,0.1,15.4")])


def test_iv_sweep_mode_start_end_steps_compliance(smu):
    mainframe = smu.parent

    smu.iv_sweep.sweep_mode(constants.SweepMode.LINEAR_TWO_WAY)
    smu.iv_sweep.sweep_range(constants.VOutputRange.MIN_2V)
    smu.iv_sweep.sweep_start(0.2)
    smu.iv_sweep.sweep_end(12.3)
    smu.iv_sweep.sweep_steps(13)
    smu.iv_sweep.current_compliance(45e-3)
    smu.iv_sweep.power_compliance(0.2)

    mainframe.write.assert_has_calls([call('WV 1,3,0,0.0,0.0,1'),
                                      call('WV 1,3,20,0.0,0.0,1'),
                                      call('WV 1,3,20,0.2,0.0,1'),
                                      call('WV 1,3,20,0.2,12.3,1'),
                                      call('WV 1,3,20,0.2,12.3,13'),
                                      call('WV 1,3,20,0.2,12.3,13,0.045'),
                                      call('WV 1,3,20,0.2,12.3,13,0.045,0.2')]
                                     )


def test_set_sweep_auto_abort(smu):
    mainframe = smu.parent

    smu.iv_sweep.sweep_auto_abort(constants.Abort.ENABLED)

    mainframe.write.assert_called_once_with("WM 2")


def test_get_sweep_auto_abort(smu):
    mainframe = smu.parent

    mainframe.ask.return_value = "WM2,2;WT1.0,0.0,0.0,0.0,0.0;"
    condition = smu.iv_sweep.sweep_auto_abort()
    assert condition == constants.Abort.ENABLED


def test_set_post_sweep_voltage_cond(smu):
    mainframe = smu.parent
    mainframe.ask.return_value = "WM2,2;WT1.0,0.0,0.0,0.0,0.0"
    smu.iv_sweep.post_sweep_voltage_condition(constants.WMDCV.Post.STOP)

    mainframe.write.assert_called_once_with("WM 2,2")


def test_get_post_sweep_voltage_cond(smu):
    mainframe = smu.parent

    mainframe.ask.return_value = "WM2,2;WT1.0,0.0,0.0,0.0,0.0"
    condition = smu.iv_sweep.post_sweep_voltage_condition()
    assert condition == constants.WM.Post.STOP



"""
Live plotting in Jupyter notebooks
using the nbagg backend and matplotlib
"""
from collections.abc import Mapping, Sequence
from copy import deepcopy
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import ticker
from matplotlib.transforms import Bbox
from numpy.ma import getmask, masked_invalid

import qcodes
from qcodes.data.data_array import DataArray

from .base import BasePlot


class MatPlot(BasePlot):
    """
    Plot x/y lines or x/y/z heatmap data. The first trace may be included
    in the constructor, other traces can be added with MatPlot.add()

    Args:
        *args: Sequence of data to plot. Each element will have its own subplot.
            An element can be a single array, or a sequence of arrays. In the
            latter case, all arrays will be plotted in the same subplot.

        figsize (Tuple[float, float]): (width, height) tuple in inches to pass
            to plt.figure. If not provided, figsize is determined from
            subplots shape

        interval: period in seconds between update checks

        subplots: either a sequence (args) or mapping (kwargs) to pass to
            plt.subplots. default is a single simple subplot (1, 1)
            you can use this to pass kwargs to the plt.figure constructor

        num: integer or None
            specifies the index of the matplotlib figure window to use. If None
            then open a new window

        **kwargs: passed along to MatPlot.add() to add the first data trace
    """

    # Maximum default number of subplot columns. Used to determine shape of
    # subplots when not explicitly provided
    max_subplot_columns = 3

    def __init__(self, *args, figsize=None, interval=1, subplots=None, num=None,
                 **kwargs):
        super().__init__(interval)

        if subplots is None:
            # Subplots is equal to number of args, or 1 if no args provided
            subplots = max(len(args), 1)

        self._init_plot(subplots, figsize, num=num)

        # Add data to plot if passed in args, kwargs are passed to all subplots
        for k, arg in enumerate(args):
            if isinstance(arg, Sequence):
                # Arg consists of multiple elements, add all to same subplot
                for subarg in arg:
                    self[k].add(subarg, **kwargs)
            else:
                # Arg is single element, add to subplot
                self[k].add(arg, **kwargs)

        self.tight_layout()

    def __getitem__(self, key):
        """
        Subplots can be accessed via indices.
        Args:
            key: subplot idx

        Returns:
            Subplot with idx key
        """
        return self.subplots[key]

    def _init_plot(self, subplots=None, figsize=None, num=None):
        if isinstance(subplots, Mapping):
            if figsize is None:
                figsize = (6, 4)
            self.fig, self.subplots = plt.subplots(figsize=figsize, num=num, squeeze=False, **subplots)
        else:
            # Format subplots as tuple (nrows, ncols)
            if isinstance(subplots, int):
                # self.max_subplot_columns defines the limit on how many
                # subplots can be in one row. Adjust subplot rows and columns
                #  accordingly
                nrows = int(np.ceil(subplots / self.max_subplot_columns))
                ncols = min(subplots, self.max_subplot_columns)
                subplots = (nrows, ncols)
            if subplots is None:
                subplots = (1,1)
            if figsize is None:
                # Adjust figsize depending on rows and columns in subplots
                figsize = self.default_figsize(subplots)

            self.fig, self.subplots = plt.subplots(*subplots, num=num,
                                                   figsize=figsize,
                                                   squeeze=False)

        # squeeze=False ensures that subplots is always a 2D array independent
        # of the number of subplots.
        # However the qcodes api assumes that subplots is always a 1D array
        # so flatten here
        self.subplots = self.subplots.flatten()

        for k, subplot in enumerate(self.subplots):
            # Include `add` method to subplots, making it easier to add data to
            # subplots. Note that subplot kwarg is 1-based, to adhere to
            # Matplotlib standards
            subplot.add = partial(self.add, subplot=k+1)

        self.title = self.fig.suptitle('')

    def clear(self, subplots=None, figsize=None):
        """
        Clears the plot window and removes all subplots and traces
        so that the window can be reused.
        """
        self.traces = []
        self.fig.clf()
        self._init_plot(subplots, figsize, num=self.fig.number)

    def add_to_plot(self, use_offset: bool=False, **kwargs):
        """
        adds one trace to this MatPlot.

        Args:
            use_offset: Whether or not ticks can have an offset
            **kwargs: with the exceptions given in the notes below
                (mostly the data!), these are passed directly to
                the matplotlib plotting routine.

        Returns:
            Plot handle for trace

        Notes:
            The following special cases apply for kwargs that are
            not passed directly to the plotting routine.

            * `subplot`: the 1-based axes number to append to (default 1)
            * if kwargs include `z`, we will draw a heatmap (ax.pcolormesh)
              `x`, `y`, and `z` are passed as positional args to pcolormesh
            * without `z` we draw a scatter/lines plot (ax.plot)
              `x`, `y`, and `fmt` (if present) are passed as positional
              args
        """
        # TODO some way to specify overlaid axes?
        # Note that there is a conversion from subplot kwarg, which is
        # 1-based, to subplot idx, which is 0-based.
        ax = self[kwargs.get('subplot', 1) - 1]
        if 'z' in kwargs:
            plot_object = self._draw_pcolormesh(ax, **kwargs)
        else:
            plot_object = self._draw_plot(ax, **kwargs)

        # Specify if axes ticks can have offset or not
        ax.ticklabel_format(useOffset=use_offset)

        self._update_labels(ax, kwargs)
        prev_default_title = self.get_default_title()

        self.traces.append({
            'config': kwargs,
            'plot_object': plot_object
        })

        if prev_default_title == self.title.get_text():
            # in case the user has updated title, don't change it anymore
            self.title.set_text(self.get_default_title())

        return plot_object

    def _update_labels(self, ax, config):
        for axletter in ("x", "y"):
            if axletter+'label' in config:
                label = config[axletter+'label']
            else:
                label = None

            # find if any kwarg from plot.add in the base class
            # matches xunit or yunit, signaling a custom unit
            if axletter+'unit' in config:
                unit = config[axletter+'unit']
            else:
                unit = None

            #  find ( more hope to) unit and label from
            # the data array inside the config
            getter = getattr(ax, f"get_{axletter}label")
            if axletter in config and not getter():
                # now if we did not have any kwarg for label or unit
                # fallback to the data_array
                if unit is None:
                    _, unit = self.get_label(config[axletter])
                if label is None:
                    label, _ = self.get_label(config[axletter])
            elif getter():
                # The axis already has label. Assume that is correct
                # We should probably check consistent units and error or warn
                # if not consistent. It's also not at all clear how to handle
                # labels/names as these will in general not be consistent on
                # at least one axis
                return
            axsetter = getattr(ax, f"set_{axletter}label")
            axsetter(f"{label} ({unit})")

    @staticmethod
    def default_figsize(subplots):
        """
        Provides default figsize for given subplots.
        Args:
            subplots (Tuple[Int, Int]): shape (nrows, ncols) of subplots

        Returns:
            Tuple[float, float]: (width, height) of default figsize
              for given subplot shape
        """
        if not isinstance(subplots, tuple):
            raise TypeError(f'Subplots {subplots} must be a tuple')
        return (3 + 3 * subplots[1], 1 + 3 * subplots[0])

    def update_plot(self):
        """
        update the plot. The DataSets themselves have already been updated
        in update, here we just push the changes to the plot.
        """
        # matplotlib doesn't know how to autoscale to a pcolormesh after the
        # first draw (relim ignores it...) so we have to do this ourselves
        bboxes = dict(zip(self.subplots, [[] for p in self.subplots]))

        for trace in self.traces:
            config = trace['config']
            plot_object = trace['plot_object']
            if 'z' in config:
                # pcolormesh doesn't seem to allow editing x and y data, only z
                # so instead, we'll remove and re-add the data.
                if plot_object:
                    plot_object.remove()

                ax = self[config.get('subplot', 1) - 1]
                kwargs = deepcopy(config)
                # figsize may be passed in as part of config.
                # pcolormesh will raise an error if this is passed to it
                # so strip it here.
                if 'figsize' in kwargs:
                    kwargs.pop('figsize')
                plot_object = self._draw_pcolormesh(ax, **kwargs)
                trace['plot_object'] = plot_object

                if plot_object:
                    bboxes[plot_object.axes].append(
                        plot_object.get_datalim(plot_object.axes.transData))
            else:
                for axletter in 'xy':
                    setter = 'set_' + axletter + 'data'
                    if axletter in config:
                        getattr(plot_object, setter)(config[axletter])

        for ax in self.subplots:
            if ax.get_autoscale_on():
                ax.relim()
                if bboxes[ax]:
                    bbox = Bbox.union(bboxes[ax])
                    if np.all(np.isfinite(ax.dataLim)):
                        # should take care of the case of lines + heatmaps
                        # where there's already a finite dataLim from relim
                        ax.dataLim.set(Bbox.union(ax.dataLim, bbox))
                    else:
                        # when there's only a heatmap, relim gives inf bounds
                        # so just completely overwrite it
                        ax.dataLim = bbox
                ax.autoscale()

        self.fig.canvas.draw()

    def _draw_plot(self, ax, y, x=None, fmt=None, subplot=1,
                   xlabel=None,
                   ylabel=None,
                   zlabel=None,
                   xunit=None,
                   yunit=None,
                   zunit=None,
                   **kwargs):
        # NOTE(alexj)stripping out subplot because which subplot we're in is
        # already described by ax, and it's not a kwarg to matplotlib's ax.plot.
        # But I didn't want to strip it out of kwargs earlier because it should
        # stay part of trace['config'].
        args = [arg for arg in [x, y, fmt] if arg is not None]

        line, = ax.plot(*args, **kwargs)
        return line

    @staticmethod
    def _make_args_for_pcolormesh(args_masked, x, y):
        """
        Make args for pcolormesh.
        pcolormesh accepts as args either
        C - a (potentially) masked array
        or
        x, y, C where x and y are the colour box edge arrays and
        are NOT allowed to be masked
        """

        if x is not None and y is not None:
            # If x and y are provided, modify the arrays such that they
            # correspond to grid corners instead of grid centers.
            # This is to ensure that pcolormesh centers correctly and
            # does not ignore edge points.
            args = []
            for k, arr in enumerate(args_masked[:-1]):
                # If a two-dimensional array is provided, only consider the
                # first row/column, depending on the axis
                if arr.ndim > 1:
                    arr = arr[0] if k == 0 else arr[:, 0]
                # first extrapolate to fill any empty values Matplotlib 2.2 no
                # longer support nans in x and y for pcolormesh.
                if np.ma.is_masked(arr[1]):
                    step_size = 1.
                    # Only the first element is not nan. We have to guess the
                    # step size
                else:
                    # the average stepsize is our best guess
                    step_size = np.ma.average(np.ma.diff(arr))

                last_good_value = arr[np.logical_not(arr.mask)][-1]
                extrapolation_start = last_good_value+step_size
                n_invalid = np.sum(arr.mask)
                extrapolation_stop = extrapolation_start+step_size*(n_invalid-1)
                # numpy (1.14) has a deprecation warning related to shared
                # masks. Let's silence this by making sure that this is
                # not shared before modifying the mask
                arr.unshare_mask()
                arr[arr.mask] = np.linspace(extrapolation_start,
                                            extrapolation_stop,
                                            num=n_invalid)
                # Now shift array to get edge coordinates rather than
                # centre coordinates
                arr_shift = np.insert(arr, 0, arr[0])
                arr_shift[0] -= step_size/2
                arr_shift[1:] += step_size/2
                args.append(arr_shift)
            args.append(args_masked[-1])
        else:
            # Only the masked value of z is used as a mask
            args = args_masked[-1:]

        return args

    def _draw_pcolormesh(self, ax, z, x=None, y=None, subplot=1,
                         xlabel=None,
                         ylabel=None,
                         zlabel=None,
                         xunit=None,
                         yunit=None,
                         zunit=None,
                         nticks=None,
                         **kwargs):
        # NOTE(alexj)stripping out subplot because which subplot we're in is already
        # described by ax, and it's not a kwarg to matplotlib's ax.plot. But I
        # didn't want to strip it out of kwargs earlier because it should stay
        # part of trace['config'].

        args_masked = [masked_invalid(arg) for arg in [x, y, z]
                       if arg is not None]

        if np.any([np.all(getmask(arg)) for arg in args_masked]):
            # if the z array is masked, don't draw at all
            # there's nothing to draw, and anyway it throws a warning
            # pcolormesh does not accept masked x and y axes, so we do not need
            # to check for them.
            return False
        if 'cmap' not in kwargs:
            kwargs['cmap'] = qcodes.config['gui']['defaultcolormap']

        args = self._make_args_for_pcolormesh(args_masked, x, y)

        pc = ax.pcolormesh(*args, **kwargs)

        # Set x and y limits if arrays are provided
        if x is not None and y is not None:
            ax.set_xlim(np.nanmin(args[0]), np.nanmax(args[0]))
            ax.set_ylim(np.nanmin(args[1]), np.nanmax(args[1]))

        # Specify preferred number of ticks with labels
        if nticks and ax.get_xscale() != 'log' and ax.get_yscale != 'log':
            ax.locator_params(nbins=nticks)

        if getattr(ax, 'qcodes_colorbar', None):
            # update_normal doesn't seem to work...
            ax.qcodes_colorbar.update_bruteforce(pc)
        else:
            # TODO: what if there are several colormeshes on this subplot,
            # do they get the same colorscale?
            # We should make sure they do, and have it include
            # the full range of both.
            ax.qcodes_colorbar = self.fig.colorbar(pc, ax=ax)

            # ideally this should have been in _update_labels, but
            # the colorbar doesn't necessarily exist there.
            # I guess we could create the colorbar no matter what,
            # and just give it a dummy mappable to start, so we could
            # put this where it belongs.
            if zunit is None:
                _, zunit = self.get_label(z)
            if zlabel is None:
                zlabel, _ = self.get_label(z)

            label = f"{zlabel} ({zunit})"
            ax.qcodes_colorbar.set_label(label)

        # Scale colors if z has elements
        cmin = np.nanmin(args_masked[-1])
        cmax = np.nanmax(args_masked[-1])
        ax.qcodes_colorbar.mappable.set_clim(cmin, cmax)

        return pc

    def save(self, filename=None):
        """
        Save current plot to filename, by default
        to the location corresponding to the default
        title.

        Args:
            filename (Optional[str]): Location of the file
        """
        default = f"{self.get_default_title()}.png"
        filename = filename or default
        self.fig.savefig(filename)

    def tight_layout(self):
        """
        Perform a tight layout on the figure. A bit of additional spacing at
        the top is also added for the title.
        """
        self.fig.tight_layout(rect=[0, 0, 1, 0.95])

    def rescale_axis(self):
        """
        Rescale axis and units for axis that are in standard units
        i.e. V, s J ... to m μ, m
        This scales units defined in BasePlot.standardunits only
        to avoid prefixes on combined or non standard units
        """
        def scale_formatter(i, pos, scale):
            return f"{i * scale:g}"

        for i, subplot in enumerate(self.subplots):
            traces = [trace for trace in self.traces if trace['config'].get('subplot', None) == i+1]
            if not traces:
                continue
            else:
                # TODO: include all traces when calculating maxval etc.
                trace = traces[0]
            for axis in 'x', 'y', 'z':
                if axis in trace['config'] and isinstance(trace['config'][axis], DataArray):
                    unit = trace['config'][axis].unit
                    label = trace['config'][axis].label
                    maxval = np.nanmax(abs(trace['config'][axis].ndarray))
                    units_to_scale = self.standardunits

                    # allow values up to a <1000. i.e. nV is used up to 1000 nV
                    prefixes = ['n', 'μ', 'm', '', 'k', 'M', 'G']
                    thresholds = [10**(-6 + 3*n) for n in range(len(prefixes))]
                    scales = [10**(9 - 3*n) for n in range(len(prefixes))]

                    if unit in units_to_scale:
                        scale = 1
                        new_unit = unit
                        for prefix, threshold, trialscale in zip(prefixes,
                                                                 thresholds,
                                                                 scales):
                            if maxval < threshold:
                                scale = trialscale
                                new_unit = prefix + unit
                                break
                        # special case the largest
                        if maxval > thresholds[-1]:
                            scale = scales[-1]
                            new_unit = prefixes[-1] + unit

                        tx = ticker.FuncFormatter(
                            partial(scale_formatter, scale=scale))
                        new_label = f"{label} ({new_unit})"
                        if axis in ('x', 'y'):
                            getattr(subplot,
                                    f"{axis}axis").set_major_formatter(
                                tx)
                            getattr(subplot, f"set_{axis}label")(
                                new_label)
                        else:
                            subplot.qcodes_colorbar.formatter = tx
                            subplot.qcodes_colorbar.set_label(new_label)
                            subplot.qcodes_colorbar.update_ticks()



import pytest
from qcodes.utils.validators import Dict


def test_dict():
    d = Dict()
    my_dict = {}
    d.validate(my_dict)
    my_int = 5
    with pytest.raises(TypeError):
        d.validate(my_int)


def test_valid_values():
    val = Dict()
    for vval in val.valid_values:
        val.validate(vval)



import logging
import select
import socket
import sys
if sys.platform == "win32":
    from msvcrt import kbhit, getch
else:
    raise RuntimeError("Dynacool server only supported on Windows")

from qcodes.instrument_drivers.QuantumDesign.\
    DynaCoolPPMS.private.commandhandler import CommandHandler

assert sys.platform == 'win32'

command_handler = CommandHandler()

log = logging.getLogger(__name__)

RECEIVE_BUFFER_SIZE = 4096
PORT = 5000
ADDRESS = ''
LINE_TERM = '\r\n'

if __name__ == '__main__':

    server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server_socket.bind((ADDRESS, PORT))
    server_socket.listen(10)

    # Dictionary to keep track of sockets and addresses.
    # Keys are sockets and values are addresses.
    # Add server socket to the dictionary first.
    socket_dict = {server_socket: (ADDRESS, PORT)}

    print(f'Server started on port {PORT}.')
    print('Press ESC to exit.')

    keep_going = True
    cmd_buffer: str = ''

    while keep_going:
        # Get the list sockets which are ready to be read through select
        # 1 second timeout so that we can process keyboard events
        read_sockets = select.select(list(socket_dict.keys()), [], [], 1)[0]

        # Keyboard
        if kbhit():
            if ord(getch()) == 27:
                print('Server exiting')
                break

        for sock in read_sockets:
            # New connection
            if sock == server_socket:
                sock_fd, address = server_socket.accept()
                socket_dict[sock_fd] = address
                print('Client ({}, {}) connected.'.format(*address))
                log.info('Client ({}, {}) connected.'.format(*address))

            # Incoming message from existing connection
            else:
                data = sock.recv(RECEIVE_BUFFER_SIZE).decode('utf-8')
                log.debug(f'Received data: {data}')
                if data:
                    data = data.replace('\n', '\r')
                    data = data.replace('\r\r', '\r')
                    cmd_buffer += data
            idx = cmd_buffer.find('\r')

            if idx >= 0:
                command = cmd_buffer[:idx].upper().strip(' ')
                cmd_buffer = cmd_buffer[idx+1:]
                if command == 'EXIT':
                    # send an error code, since the driver expects that for all
                    # commands
                    sock.send(b'0')
                    print('Server exiting.')
                    keep_going = False
                elif command == 'CLOSE':
                    # send an error code, since the driver expects that for all
                    # commands
                    sock.send(b'0')
                    print('Client ({}, {}) disconnected.'.format(*socket_dict[sock]))
                    log.info('Client ({}, {}) disconnected.'.format(*socket_dict[sock]))
                    socket_dict.pop(sock, None)
                    sock.close()
                else:
                    response = command_handler(command)
                    sock.send(bytes(f'{response}{LINE_TERM}', 'utf-8'))


    server_socket.close()



# a driver to interact with the dummy.yaml simulated instrument
#
# dummy = Dummy('dummy', 'GPIB::8::INSTR', visalib='dummy.yaml@sim',
#               terminator='\n', device_clear=False)
from typing import Any

from qcodes.instrument.visa import VisaInstrument
import qcodes.utils.validators as vals


class Dummy(VisaInstrument):

    def __init__(self, name: str, address: str, **kwargs: Any):
        super().__init__(name, address, **kwargs)

        self.connect_message()

        self.add_parameter('frequency',
                           set_cmd='FREQ {}',
                           get_cmd='FREQ?',
                           unit='Hz',
                           get_parser=float,
                           vals=vals.Numbers(10, 1000))



from deepdiff import DeepDiff

from qcodes.dataset.descriptions.rundescriber import RunDescriber
from qcodes.dataset.descriptions.versioning.converters import (new_to_old,
                                                               old_to_new,
                                                               v0_to_v1,
                                                               v0_to_v2,
                                                               v1_to_v0,
                                                               v1_to_v2,
                                                               v2_to_v0,
                                                               v2_to_v1)
from qcodes.dataset.descriptions.versioning.rundescribertypes import (
    RunDescriberV0Dict, RunDescriberV1Dict, RunDescriberV2Dict,
    RunDescriberV3Dict)
from qcodes.dataset.descriptions.versioning.serialization import \
    from_dict_to_current
from qcodes.dataset.descriptions.versioning.v0 import InterDependencies


def test_convert_v0_to_newer(some_paramspecs):
    pgroup1 = some_paramspecs[1]

    interdeps = InterDependencies(pgroup1['ps1'],
                                  pgroup1['ps2'],
                                  pgroup1['ps3'],
                                  pgroup1['ps4'],
                                  pgroup1['ps6'])

    v0 = RunDescriberV0Dict(interdependencies=interdeps._to_dict(), version=0)
    v1 = v0_to_v1(v0)
    v2 = v0_to_v2(v0)

    _assert_dicts_are_related_as_expected(v0, v1, v2)


def test_convert_v1(some_interdeps):
    interdeps_ = some_interdeps[0]

    v1 = RunDescriberV1Dict(interdependencies=interdeps_._to_dict(),
                            version=1)
    v0 = v1_to_v0(v1)
    v2 = v1_to_v2(v1)
    _assert_dicts_are_related_as_expected(v0, v1, v2)


def test_convert_v2(some_interdeps):
    interdeps_ = some_interdeps[0]
    interdeps = new_to_old(interdeps_)

    v2 = RunDescriberV2Dict(interdependencies=interdeps._to_dict(),
                            interdependencies_=interdeps_._to_dict(),
                            version=2)
    v1 = v2_to_v1(v2)
    v0 = v2_to_v0(v2)
    _assert_dicts_are_related_as_expected(v0, v1, v2)


def _assert_dicts_are_related_as_expected(v0, v1, v2):
    assert v1['interdependencies'] == old_to_new(
        InterDependencies._from_dict(v0['interdependencies'])
    )._to_dict()
    assert v1['version'] == 1
    assert len(v1) == 2

    # conversion does not preserve order in the dict so use deepdiff to compare
    assert DeepDiff(v2['interdependencies'], v0['interdependencies'],
                    ignore_order=True) == {}
    assert v2['interdependencies_'] == v1['interdependencies']
    assert v2['version'] == 2
    assert len(v2) == 3


def test_construct_current_rundescriber_from_v0(some_paramspecs):

    pgroup1 = some_paramspecs[1]

    interdeps = InterDependencies(pgroup1['ps1'],
                                  pgroup1['ps2'],
                                  pgroup1['ps3'],
                                  pgroup1['ps4'],
                                  pgroup1['ps6'])
    v0 = RunDescriberV0Dict(interdependencies=interdeps._to_dict(), version=0)
    rds1 = RunDescriber._from_dict(v0)

    rds_upgraded = from_dict_to_current(v0)

    expected_v3_dict = RunDescriberV3Dict(
        interdependencies=interdeps._to_dict(),
        interdependencies_=old_to_new(interdeps)._to_dict(),
        version=3,
        shapes=None,
    )
    assert DeepDiff(rds1._to_dict(), expected_v3_dict,
                    ignore_order=True) == {}
    assert DeepDiff(rds_upgraded._to_dict(), expected_v3_dict,
                    ignore_order=True) == {}


def test_construct_current_rundescriber_from_v1(some_interdeps):
    interdeps_ = some_interdeps[0]
    interdeps = new_to_old(interdeps_)

    v1 = RunDescriberV1Dict(interdependencies=interdeps_._to_dict(),
                            version=1)
    rds1 = RunDescriber._from_dict(v1)
    rds_upgraded = from_dict_to_current(v1)

    expected_v3_dict = RunDescriberV3Dict(
        interdependencies=interdeps._to_dict(),
        interdependencies_=interdeps_._to_dict(),
        version=3,
        shapes=None,
    )
    assert rds1._to_dict() == expected_v3_dict
    assert rds_upgraded._to_dict() == expected_v3_dict


def test_construct_current_rundescriber_from_v2(some_interdeps):
    interdeps_ = some_interdeps[0]
    interdeps = new_to_old(interdeps_)

    v2 = RunDescriberV2Dict(interdependencies=interdeps._to_dict(),
                            interdependencies_=interdeps_._to_dict(),
                            version=2)

    expected_v3_dict = RunDescriberV3Dict(
        interdependencies=interdeps._to_dict(),
        interdependencies_=interdeps_._to_dict(),
        version=3,
        shapes=None,
    )
    rds1 = RunDescriber._from_dict(v2)
    rds_upgraded = from_dict_to_current(v2)

    assert rds1._to_dict() == expected_v3_dict
    assert rds_upgraded._to_dict() == expected_v3_dict


def test_construct_current_rundescriber_from_v3(some_interdeps):
    interdeps_ = some_interdeps[0]
    interdeps = new_to_old(interdeps_)

    v3 = RunDescriberV3Dict(interdependencies=interdeps._to_dict(),
                            interdependencies_=interdeps_._to_dict(),
                            version=3,
                            shapes=None)
    rds1 = RunDescriber._from_dict(v3)
    rds_upgraded = from_dict_to_current(v3)
    assert rds1._to_dict() == v3
    assert rds_upgraded._to_dict() == v3


def test_construct_current_rundescriber_from_fake_v4(some_interdeps):
    interdeps_ = some_interdeps[0]
    interdeps = new_to_old(interdeps_)

    v4 = RunDescriberV3Dict(interdependencies=interdeps._to_dict(),
                            interdependencies_=interdeps_._to_dict(),
                            version=4,
                            shapes=None)
    v4['foobar'] = {"foo": ["bar"]}
    rds1 = RunDescriber._from_dict(v4)
    rds_upgraded = from_dict_to_current(v4)
    v3 = v4.copy()
    v3.pop('foobar')
    v3['version'] = 3
    assert rds1._to_dict() == v3
    assert rds_upgraded._to_dict() == v3



"""
This module provides infrastructure for wrapping DLL libraries, loaded using
:mod:`ctypes`. With this infrastructure, one just needs to subclass
:class:`.WrappedDll` and define class methods for the functions of interest
from the DLL library with mostly python types in mind, and conveniently
specify their signatures in terms of :mod:`ctypes` types.
"""

import ctypes
import logging
from typing import Type, Dict, NamedTuple, Sequence, NewType, List, Any, TypeVar, Callable, Tuple
from threading import Lock
import concurrent
from functools import partial
from weakref import WeakValueDictionary

from qcodes.instrument.parameter import _BaseParameter
from .utils import TraceParameter
from .constants import API_SUCCESS, API_DMA_IN_PROGRESS, ERROR_CODES, \
    ReturnCode


logger = logging.getLogger(__name__)

# Define aliases for ctypes that match Alazar's notation.
RETURN_CODE = NewType('RETURN_CODE', ctypes.c_uint)


# FUNCTIONS #
T = TypeVar('T')


def _api_call_task(
        lock: Lock,
        c_func: Callable[..., int],
        callback: Callable[[], None],
        *args: Any) -> int:
    with lock:
        retval = c_func(*args)
    callback()
    return retval


def _normalize_params(*args: T) -> List[T]:
    args_out: List[T] = []
    for arg in args:
        if isinstance(arg, _BaseParameter):
            args_out.append(arg.raw_value)
        else:
            args_out.append(arg)
    return args_out


def _mark_params_as_updated(*args: Any) -> None:
    for arg in args:
        if isinstance(arg, TraceParameter):
            arg._set_updated()


def _check_error_code(
        return_code: int,
        func: Callable[..., Any],
        arguments: Tuple[Any, ...]
) -> Tuple[Any, ...]:
    if (return_code != API_SUCCESS) and (return_code != API_DMA_IN_PROGRESS):
        argrepr = repr(arguments)
        if len(argrepr) > 100:
            argrepr = argrepr[:96] + '...]'

        logger.error(f'Alazar API returned code {return_code} from function '
                     f'{func.__name__} with args {argrepr}')

        if return_code not in ERROR_CODES:
            raise RuntimeError(
                'unknown error {} from function {} with args: {}'.format(
                    return_code, func.__name__, argrepr))
        raise RuntimeError(
            'error {}: {} from function {} with args: {}'.format(
                return_code,
                ERROR_CODES[ReturnCode(return_code)],
                func.__name__,
                argrepr))

    return arguments


def _convert_bytes_to_str(
        output: bytes,
        func: Callable[..., Any],
        arguments: Tuple[Any, ...]
) -> str:
    return output.decode()


# CLASSES #


class Signature(NamedTuple):
    return_type: Type[Any] = RETURN_CODE
    argument_types: Sequence[Type[Any]] = ()


class DllWrapperMeta(type):
    """DLL-path-based 'singleton' metaclass for DLL wrapper classes"""

    # Only allow a single instance per DLL path.
    _instances: "WeakValueDictionary[str, Any]" = WeakValueDictionary()

    # Note: without the 'type: ignore' for the ``__call__`` method below, mypy
    # generates 'Signature of "__call__" incompatible with supertype "type"'
    # error, which is an indicator of Liskov principle violation - subtypes
    # should not change the method signatures, but we need it here in order to
    # use the ``dll_path`` argument which the ``type`` superclass obviously
    # does not have in its ``__call__`` method.
    def __call__(  # type: ignore[override]
            cls,
            dll_path: str,
            *args: Any,
            **kwargs: Any
    ) -> Any:
        api = cls._instances.get(dll_path, None)
        if api is not None:
            logger.debug(
                f"Using existing instance for DLL path {dll_path}.")
            return api
        else:
            logger.debug(
                f"Creating new instance for DLL path {dll_path}.")
            # strong reference:
            new_api = super().__call__(dll_path, *args, **kwargs)
            cls._instances[dll_path] = new_api
            return new_api


class WrappedDll(metaclass=DllWrapperMeta):
    """
    A base class for wrapping DLL libraries.

    Note that this class is still quite specific to Alazar ATS DLL library.

    This class uses dictionary of the :attr:`signatures` attribute in order
    to assign ``argtypes`` and ``restype`` attributes for functions of
    a loaded DLL library (from the ``_dll`` attribute of the class).
    If ``restype`` is of type ``RETURN_CODE``, then an exception is
    raised in case the return code is an Alazar error code. For string-alike
    ``restype`` s, the returned value is converted to a python string.

    Functions are executed in a single separate thread (see what the
    ``_executor`` gets initialize to), hence the class also has a lock
    instance in the ``_lock`` attribute that is used to wrap around the
    actual calls to the DLL library.

    Method ``_sync_dll_call`` is supposed to be called when a subclass
    implements calls to functions of the loaded DLL.

    Args:
        dll_path: Path to the DLL library to load and wrap
    """

    signatures: Dict[str, Signature] = {}
    """
    Signatures for loaded DLL functions;
    It is to be filled with :class:`Signature` instances for the DLL
    functions of interest in a subclass.
    """


    # This is the DLL library instance.
    _dll: ctypes.CDLL

    # This lock guards DLL calls.
    _lock: Lock

    # This executor is used to execute DLL calls.
    _executor: concurrent.futures.Executor

    def __init__(self, dll_path: str):
        super().__init__()
        self._dll = ctypes.cdll.LoadLibrary(dll_path)
        self.__apply_signatures()
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self._lock = Lock()  # ATS API DLL is not guaranteed to be thread-safe

    def __apply_signatures(self) -> None:
        """
        Apply :mod:`ctypes` signatures for all of the C library functions
        specified in :attr:`signatures` attribute.
        """
        for name, signature in self.signatures.items():
            c_func = getattr(self._dll, name)

            c_func.argtypes = signature.argument_types

            ret_type = signature.return_type
            if ret_type is RETURN_CODE:
                ret_type = ret_type.__supertype__
                c_func.errcheck = _check_error_code
            elif ret_type in (ctypes.c_char_p, ctypes.c_char,
                              ctypes.c_wchar, ctypes.c_wchar_p):
                c_func.errcheck = _convert_bytes_to_str
            c_func.restype = ret_type

    def _sync_dll_call(self, c_name: str, *args: Any) -> Any:
        """Call given function from the DLL library on the given arguments"""
        c_func = getattr(self._dll, c_name)
        future = self._executor.submit(
            _api_call_task,
            self._lock,
            c_func,
            partial(_mark_params_as_updated, *args),
            *_normalize_params(*args)
        )
        return future.result()



import logging
import time
from distutils.version import LooseVersion
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Union, cast

import numpy as np

from qcodes.instrument.channel import InstrumentChannel
from qcodes.instrument.visa import VisaInstrument
from qcodes.math_utils.field_vector import FieldVector

log = logging.getLogger(__name__)
visalog = logging.getLogger('qcodes.instrument.visa')


def _response_preparser(bare_resp: str) -> str:
    """
    Pre-parse response from the instrument
    """
    return bare_resp.replace(':', '')


def _signal_parser(our_scaling: float, response: str) -> float:
    """
    Parse a response string into a correct SI value.

    Args:
        our_scaling: Whatever scale we might need to apply to get from
            e.g. A/min to A/s.
        response: What comes back from instrument.ask
    """

    # there might be a scale before the unit. We only want to deal in SI
    # units, so we translate the scale
    scale_to_factor = {'n': 1e-9, 'u': 1e-6, 'm': 1e-3,
                       'k': 1e3, 'M': 1e6}

    numchars = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.', '-']

    response = _response_preparser(response)
    digits = "".join(d for d in response if d in numchars)
    scale_and_unit = response[len(digits) :]
    if scale_and_unit == "":
        their_scaling: float = 1
    elif scale_and_unit[0] in scale_to_factor.keys():
        their_scaling = scale_to_factor[scale_and_unit[0]]
    else:
        their_scaling = 1

    return float(digits)*their_scaling*our_scaling


class MercuryWorkerPS(InstrumentChannel):
    """
    Class to hold a worker power supply for the MercuryiPS
    """

    def __init__(self, parent: VisaInstrument, name: str, UID: str) -> None:
        """
        Args:
            parent: The Instrument instance of the MercuryiPS
            name: The 'colloquial' name of the PS
            UID: The UID as used internally by the MercuryiPS, e.g.
                'GRPX'
        """
        if ':' in UID:
            raise ValueError('Invalid UID. Must be axis group name or device '
                             'name, e.g. "GRPX" or "PSU.M1"')

        super().__init__(parent, name)
        self.uid = UID

        # The firmware update from 2.5 -> 2.6 changed the command
        # syntax slightly
        if LooseVersion(self.root_instrument.firmware) >= LooseVersion('2.6'):
            self.psu_string = "SPSU"
        else:
            self.psu_string = "PSU"

        self.add_parameter('voltage',
                           label='Output voltage',
                           get_cmd=partial(self._param_getter, 'SIG:VOLT'),
                           unit='V',
                           get_parser=partial(_signal_parser, 1))

        self.add_parameter('current',
                           label='Output current',
                           get_cmd=partial(self._param_getter, 'SIG:CURR'),
                           unit='A',
                           get_parser=partial(_signal_parser, 1))

        self.add_parameter('current_persistent',
                           label='Output persistent current',
                           get_cmd=partial(self._param_getter, 'SIG:PCUR'),
                           unit='A',
                           get_parser=partial(_signal_parser, 1))

        self.add_parameter('current_target',
                           label='Target current',
                           get_cmd=partial(self._param_getter, 'SIG:CSET'),
                           unit='A',
                           get_parser=partial(_signal_parser, 1))

        self.add_parameter('field_target',
                           label='Target field',
                           get_cmd=partial(self._param_getter, 'SIG:FSET'),
                           set_cmd=partial(self._param_setter, 'SIG:FSET'),
                           unit='T',
                           get_parser=partial(_signal_parser, 1))

        # NB: The current ramp rate follows the field ramp rate
        # (converted via the ATOB param)
        self.add_parameter('current_ramp_rate',
                           label='Ramp rate (current)',
                           unit='A/s',
                           get_cmd=partial(self._param_getter, 'SIG:RCST'),
                           get_parser=partial(_signal_parser, 1/60))

        self.add_parameter('field_ramp_rate',
                           label='Ramp rate (field)',
                           unit='T/s',
                           set_cmd=partial(self._param_setter, 'SIG:RFST'),
                           get_cmd=partial(self._param_getter, 'SIG:RFST'),
                           get_parser=partial(_signal_parser, 1/60),
                           set_parser=lambda x: x*60)

        self.add_parameter('field',
                           label='Field strength',
                           unit='T',
                           get_cmd=partial(self._param_getter, 'SIG:FLD'),
                           get_parser=partial(_signal_parser, 1))

        self.add_parameter('field_persistent',
                           label='Persistent field strength',
                           unit='T',
                           get_cmd=partial(self._param_getter, 'SIG:PFLD'),
                           get_parser=partial(_signal_parser, 1))

        self.add_parameter('ATOB',
                           label='Current to field ratio',
                           unit='A/T',
                           get_cmd=partial(self._param_getter, 'ATOB'),
                           get_parser=partial(_signal_parser, 1),
                           set_cmd=partial(self._param_setter, 'ATOB'))

        self.add_parameter('ramp_status',
                           label='Ramp status',
                           get_cmd=partial(self._param_getter, 'ACTN'),
                           set_cmd=self._ramp_status_setter,
                           get_parser=_response_preparser,
                           val_mapping={'HOLD': 'HOLD',
                                        'TO SET': 'RTOS',
                                        'CLAMP': 'CLMP',
                                        'TO ZERO': 'RTOZ'})

    def ramp_to_target(self) -> None:
        """
        Unconditionally ramp this PS to its target
        """
        status = self.ramp_status()
        if status == 'CLAMP':
            self.ramp_status('HOLD')
        self.ramp_status('TO SET')

    def _ramp_status_setter(self, cmd: str) -> None:
        status_now = self.ramp_status()
        if status_now == 'CLAMP' and cmd == 'RTOS':
            raise ValueError(f'Error in ramping unit {self.uid}: '
                             'Can not ramp to target value; power supply is '
                             'clamped. Unclamp first by setting ramp status '
                             'to HOLD.')
        else:
            partial(self._param_setter, 'ACTN')(cmd)

    def _param_getter(self, get_cmd: str) -> str:
        """
        General getter function for parameters

        Args:
            get_cmd: raw string for the command, e.g. 'SIG:VOLT'

        Returns:
            The response. Cf. MercuryiPS.ask for how much is returned
        """
        dressed_cmd = f"READ:DEV:{self.uid}:{self.psu_string}:{get_cmd}"

        resp = self._parent.ask(dressed_cmd)

        return resp

    def _param_setter(self, set_cmd: str, value: Union[float, str]) -> None:
        """
        General setter function for parameters

        Args:
            set_cmd: raw string for the command, e.g. 'SIG:FSET'
        """
        dressed_cmd = f"SET:DEV:{self.uid}:{self.psu_string}:{set_cmd}:{value}"
        # the instrument always very verbosely responds
        # the return value of `ask`
        # holds the value reported back by the instrument
        self._parent.ask(dressed_cmd)

        # TODO: we could use the opportunity to check that we did set/achieve
        #  the intended value


class MercuryiPS(VisaInstrument):
    """
    Driver class for the QCoDeS Oxford Instruments MercuryiPS magnet power
    supply
    """

    def __init__(self, name: str, address: str, visalib: Optional[str] = None,
                 field_limits: Optional[Callable[[float,
                                                  float,
                                                  float], bool]] = None,
                 **kwargs: Any) -> None:
        """
        Args:
            name: The name to give this instrument internally in QCoDeS
            address: The VISA resource of the instrument. Note that a
                socket connection to port 7020 must be made
            visalib: The VISA library to use. Leave blank if not in simulation
                mode.
            field_limits: A function describing the allowed field
                range (T). The function shall take (x, y, z) as an input and
                return a boolean describing whether that field value is
                acceptable.
        """

        if field_limits is not None and not(callable(field_limits)):
            raise ValueError('Got wrong type of field_limits. Must be a '
                             'function from (x, y, z) -> Bool. Received '
                             f'{type(field_limits)} instead.')

        if visalib:
            visabackend = visalib.split('@')[1]
        else:
            visabackend = 'NI'

        # ensure that a socket is used unless we are in simulation mode
        if not address.endswith('SOCKET') and visabackend != 'sim':
            raise ValueError('Incorrect VISA resource name. Must be of type '
                             'TCPIP0::XXX.XXX.XXX.XXX::7020::SOCKET.')

        super().__init__(name, address, terminator='\n', visalib=visalib,
                         **kwargs)

        # to ensure a correct snapshot, we must wrap the get function
        self.IDN.get = self.IDN._wrap_get(self._idn_getter)

        self.firmware = self.IDN()['firmware']

        # TODO: Query instrument to ensure which PSUs are actually present
        for grp in ['GRPX', 'GRPY', 'GRPZ']:
            psu_name = grp
            psu = MercuryWorkerPS(self, psu_name, grp)
            self.add_submodule(psu_name, psu)

        self._field_limits = (field_limits if field_limits else
                              lambda x, y, z: True)

        self._target_vector = FieldVector(x=self.GRPX.field(),
                                          y=self.GRPY.field(),
                                          z=self.GRPZ.field())

        for coord, unit in zip(
                ['x', 'y', 'z', 'r', 'theta',   'phi',     'rho'],
                ['T', 'T', 'T', 'T', 'degrees', 'degrees', 'T']):
            self.add_parameter(name=f'{coord}_target',
                               label=f'{coord.upper()} target field',
                               unit=unit,
                               get_cmd=partial(self._get_component, coord),
                               set_cmd=partial(self._set_target, coord))

            self.add_parameter(name=f'{coord}_measured',
                               label=f'{coord.upper()} measured field',
                               unit=unit,
                               get_cmd=partial(self._get_measured, [coord]))

            self.add_parameter(name=f'{coord}_ramp',
                               label=f'{coord.upper()} ramp field',
                               unit=unit,
                               docstring='A safe ramp for each coordinate',
                               get_cmd=partial(self._get_component, coord),
                               set_cmd=partial(self._set_target_and_ramp,
                                               coord, 'safe'))

            if coord in ['r', 'theta', 'phi', 'rho']:
                self.add_parameter(name=f'{coord}_simulramp',
                                   label=f'{coord.upper()} ramp field',
                                   unit=unit,
                                   docstring='A simultaneous blocking ramp for'
                                             ' a combined coordinate',
                                   get_cmd=partial(self._get_component, coord),
                                   set_cmd=partial(self._set_target_and_ramp,
                                                   coord, 'simul_block'))

        # FieldVector-valued parameters #

        self.add_parameter(name="field_target",
                           label="target field",
                           unit="T",
                           get_cmd=self._get_target_field,
                           set_cmd=self._set_target_field)

        self.add_parameter(name="field_measured",
                           label="measured field",
                           unit="T",
                           get_cmd=self._get_field)

        self.add_parameter(name="field_ramp_rate",
                           label="ramp rate",
                           unit="T/s",
                           get_cmd=self._get_ramp_rate,
                           set_cmd=self._set_ramp_rate)

        self.connect_message()

    def _get_component(self, coordinate: str) -> float:
        return self._target_vector.get_components(coordinate)[0]

    def _get_target_field(self) -> FieldVector:
        return FieldVector(
            **{
                coord: self._get_component(coord)
                for coord in 'xyz'
            }
        )

    def _get_ramp_rate(self) -> FieldVector:
        return FieldVector(
            x=self.GRPX.field_ramp_rate(),
            y=self.GRPY.field_ramp_rate(),
            z=self.GRPZ.field_ramp_rate(),
        )

    def _set_ramp_rate(self, rate: FieldVector) -> None:
        self.GRPX.field_ramp_rate(rate.x)
        self.GRPY.field_ramp_rate(rate.y)
        self.GRPZ.field_ramp_rate(rate.z)

    def _get_measured(self, coordinates: List[str]) -> Union[float,
                                                             List[float]]:
        """
        Get the measured value of a coordinate. Measures all three fields
        and computes whatever coordinate we asked for.
        """
        meas_field = FieldVector(x=self.GRPX.field(),
                                 y=self.GRPY.field(),
                                 z=self.GRPZ.field())

        if len(coordinates) == 1:
            return meas_field.get_components(*coordinates)[0]
        else:
            return meas_field.get_components(*coordinates)

    def _get_field(self) -> FieldVector:
        return FieldVector(
            x=self.x_measured(),
            y=self.y_measured(),
            z=self.z_measured()
        )

    def _set_target(self, coordinate: str, target: float) -> None:
        """
        The function to set a target value for a coordinate, i.e. the set_cmd
        for the XXX_target parameters
        """
        # first validate the new target
        valid_vec = FieldVector()
        valid_vec.copy(self._target_vector)
        valid_vec.set_component(**{coordinate: target})
        components = valid_vec.get_components('x', 'y', 'z')
        if not self._field_limits(*components):
            raise ValueError(f'Cannot set {coordinate} target to {target}, '
                             'that would violate the field_limits. ')

        # update our internal target cache
        self._target_vector.set_component(**{coordinate: target})

        # actually assign the target on the workers
        cartesian_targ = self._target_vector.get_components('x', 'y', 'z')
        for targ, worker in zip(cartesian_targ, self.submodules.values()):
            if not isinstance(worker, MercuryWorkerPS):
                raise RuntimeError(f"Expected a MercuryWorkerPS but got "
                                   f"{type(worker)}")
            worker.field_target(targ)

    def _set_target_field(self, field: FieldVector) -> None:
        for coord in 'xyz':
            self._set_target(coord, field[coord])

    def _idn_getter(self) -> Dict[str, str]:
        """
        Parse the raw non-SCPI compliant IDN string into an IDN dict

        Returns:
            The normal IDN dict
        """
        raw_idn_string = self.ask('*IDN?')
        resps = raw_idn_string.split(':')

        idn_dict = {'model': resps[2], 'vendor': resps[1],
                    'serial': resps[3], 'firmware': resps[4]}

        return idn_dict

    def _ramp_simultaneously(self) -> None:
        """
        Ramp all three fields to their target simultaneously at their given
        ramp rates. NOTE: there is NO guarantee that this does not take you
        out of your safe region. Use with care.
        """
        for worker in self.submodules.values():
            if not isinstance(worker, MercuryWorkerPS):
                raise RuntimeError(f"Expected a MercuryWorkerPS but got "
                                   f"{type(worker)}")
            worker.ramp_to_target()

    def _ramp_simultaneously_blocking(self) -> None:
        """
        Ramp all three fields to their target simultaneously at their given
        ramp rates. NOTE: there is NO guarantee that this does not take you
        out of your safe region. Use with care. This function is BLOCKING.
        """
        self._ramp_simultaneously()

        for worker in self.submodules.values():
            if not isinstance(worker, MercuryWorkerPS):
                raise RuntimeError(f"Expected a MercuryWorkerPS but got "
                                   f"{type(worker)}")
            # wait for the ramp to finish, we don't care about the order
            while worker.ramp_status() == 'TO SET':
                time.sleep(0.1)

        self.update_field()

    def _ramp_safely(self) -> None:
        """
        Ramp all three fields to their target using the 'first-down-then-up'
        sequential ramping procedure. This function is BLOCKING.
        """
        meas_vals = self._get_measured(['x', 'y', 'z'])
        targ_vals = self._target_vector.get_components('x', 'y', 'z')
        order = np.argsort(np.abs(np.array(targ_vals) - np.array(meas_vals)))

        for worker in np.array(list(self.submodules.values()))[order]:
            worker.ramp_to_target()
            # now just wait for the ramp to finish
            # (unless we are testing)
            if self.visabackend == 'sim':
                pass
            else:
                while worker.ramp_status() == 'TO SET':
                    time.sleep(0.1)

        self.update_field()

    def update_field(self) -> None:
        """
        Update all the field components.
        """
        coords = ['x', 'y', 'z', 'r', 'theta', 'phi', 'rho']
        _ = self._get_field()
        [getattr(self, f'{i}_measured').get() for i in coords]

    def is_ramping(self) -> bool:
        """
        Returns True if any axis has a ramp status that is either 'TO SET' or
        'TO ZERO'
        """
        ramping_statuus = ['TO SET', 'TO ZERO']
        is_x_ramping = self.GRPX.ramp_status() in ramping_statuus
        is_y_ramping = self.GRPY.ramp_status() in ramping_statuus
        is_z_ramping = self.GRPZ.ramp_status() in ramping_statuus

        return is_x_ramping or is_y_ramping or is_z_ramping

    def set_new_field_limits(self, limit_func: Callable[[float,
                                                         float,
                                                         float], bool]) -> None:
        """
        Assign a new field limit function to the driver

        Args:
            limit_func: must be a function mapping (Bx, By, Bz) -> True/False
              where True means that the field is INSIDE the allowed region
        """

        # first check that the current target is allowed
        if not limit_func(*self._target_vector.get_components('x', 'y', 'z')):
            raise ValueError('Can not assign new limit function; present '
                             'target is illegal. Please change the target '
                             'and try again.')

        self._field_limits = limit_func

    def ramp(self, mode: str = "safe") -> None:
        """
        Ramp the fields to their present target value

        Args:
            mode: how to ramp, either 'simul', 'simul-block' or 'safe'. In
              'simul' and 'simul-block' mode, the fields are ramping
              simultaneously in a non-blocking mode and blocking mode,
              respectively. There is no safety check that the safe zone is not
              exceeded. In 'safe' mode, the fields are ramped one-by-one in a
              blocking way that ensures that the total field stays within the
              safe region (provided that this region is convex).
        """
        if mode not in ['simul', 'safe', 'simul_block']:
            raise ValueError('Invalid ramp mode. Please provide either "simul"'
                             ', "safe" or "simul_block".')

        meas_vals = self._get_measured(['x', 'y', 'z'])
        # we asked for three coordinates, so we know that we got a list
        meas_vals = cast(List[float], meas_vals)

        for cur, worker in zip(meas_vals, self.submodules.values()):
            if not isinstance(worker, MercuryWorkerPS):
                raise RuntimeError(f"Expected a MercuryWorkerPS but got "
                                   f"{type(worker)}")
            if worker.field_target() != cur:
                if worker.field_ramp_rate() == 0:
                    raise ValueError(f'Can not ramp {worker}; ramp rate set to'
                                     ' zero!')

        # then the actual ramp
        {'simul': self._ramp_simultaneously,
         'safe': self._ramp_safely,
         'simul_block': self._ramp_simultaneously_blocking}[mode]()

    def _set_target_and_ramp(self, coordinate: str, mode: str, target: float) -> None:
        """Convenient method to combine setting target and ramping"""
        self._set_target(coordinate, target)
        self.ramp(mode)

    def ask(self, cmd: str) -> str:
        """
        Since Oxford Instruments implement their own version of a SCPI-like
        language, we implement our own reader. Note that this command is used
        for getting and setting (asking and writing) alike.

        Args:
            cmd: the command to send to the instrument
        """

        visalog.debug(f"Writing to instrument {self.name}: {cmd}")
        resp = self.visa_handle.query(cmd)
        visalog.debug(f"Got instrument response: {resp}")

        if 'INVALID' in resp:
            log.error(f'Invalid command. Got response: {resp}')
            base_resp = resp
        # if the command was not invalid, it can either be a SET or a READ
        # SET:
        elif resp.endswith('VALID'):
            base_resp = resp.split(':')[-2]
        # READ:
        else:
            # For "normal" commands only (e.g. '*IDN?' is excepted):
            # the response of a valid command echoes back said command,
            # thus we remove that part
            base_cmd = cmd.replace('READ:', '')
            base_resp = resp.replace(f'STAT:{base_cmd}', '')

        return base_resp



from typing import Any

from .ATS import AlazarTech_ATS
from .utils import TraceParameter
from qcodes.utils import validators


class AlazarTech_ATS9440(AlazarTech_ATS):
    """
    This class is the driver for the ATS9440 board
    it inherits from the ATS base class
    """
    samples_divisor = 32
    channels = 4

    def __init__(self, name: str,
                 dll_path: str = 'C:\\WINDOWS\\System32\\ATSApi.dll',
                 **kwargs: Any):
        super().__init__(name, dll_path=dll_path, **kwargs)

        # add parameters

        # ----- Parameters for the configuration of the board -----
        self.add_parameter(name='clock_source',
                           parameter_class=TraceParameter,
                           label='Clock Source',
                           unit=None,
                           initial_value='INTERNAL_CLOCK',
                           val_mapping={'INTERNAL_CLOCK': 1,
                                        'FAST_EXTERNAL_CLOCK': 2,
                                        'SLOW_EXTERNAL_CLOCK': 4,
                                        'EXTERNAL_CLOCK_10MHz_REF': 7})
        self.add_parameter(name='external_sample_rate',
                           parameter_class=TraceParameter,
                           label='External Sample Rate',
                           unit='S/s',
                           vals=validators.MultiType(validators.Ints(1000000, 125000000),
                                                     validators.Enum('UNDEFINED')),
                           initial_value='UNDEFINED')
        self.add_parameter(name='sample_rate',
                           parameter_class=TraceParameter,
                           label='Internal Sample Rate',
                           unit='S/s',
                           initial_value=100000000,
                           val_mapping={1_000: 1,
                                        2_000: 2,
                                        5_000: 4,
                                       10_000: 8,
                                       20_000: 10,
                                       50_000: 12,
                                      100_000: 14,
                                      200_000: 16,
                                      500_000: 18,
                                    1_000_000: 20,
                                    2_000_000: 24,
                                    5_000_000: 26,
                                   10_000_000: 28,
                                   20_000_000: 30,
                                   50_000_000: 34,
                                  100_000_000: 36,
                                  125_000_000: 38,
                             'EXTERNAL_CLOCK': 64,
                                  'UNDEFINED': 'UNDEFINED'})
        self.add_parameter(name='clock_edge',
                           parameter_class=TraceParameter,
                           label='Clock Edge',
                           unit=None,
                           initial_value='CLOCK_EDGE_RISING',
                           val_mapping={'CLOCK_EDGE_RISING': 0,
                                        'CLOCK_EDGE_FALLING': 1})
        self.add_parameter(name='decimation',
                           parameter_class=TraceParameter,
                           label='Decimation',
                           unit=None,
                           initial_value=1,
                           vals=validators.Ints(1, 100000))
        for i in range(1, self.channels+1):
            self.add_parameter(name=f'coupling{i}',
                               parameter_class=TraceParameter,
                               label=f'Coupling channel {i}',
                               unit=None,
                               initial_value='DC',
                               val_mapping={'AC': 1, 'DC': 2})
            self.add_parameter(name=f'channel_range{i}',
                               parameter_class=TraceParameter,
                               label=f'Range channel {i}',
                               unit='V',
                               initial_value=0.1,
                               val_mapping={0.1: 5,
                                            0.2: 6,
                                            0.4: 7,
                                            1: 10,
                                            2: 11,
                                            4: 12})
            self.add_parameter(name=f'impedance{i}',
                               parameter_class=TraceParameter,
                               label=f'Impedance channel {i}',
                               unit='Ohm',
                               initial_value=50,
                               val_mapping={50: 2})

            self.add_parameter(name=f'bwlimit{i}',
                               parameter_class=TraceParameter,
                               label=f'Bandwidth limit channel {i}',
                               unit=None,
                               initial_value='DISABLED',
                               val_mapping={'DISABLED': 0,
                                            'ENABLED': 1})
        self.add_parameter(name='trigger_operation',
                           parameter_class=TraceParameter,
                           label='Trigger Operation',
                           unit=None,
                           initial_value='TRIG_ENGINE_OP_J',
                           val_mapping={'TRIG_ENGINE_OP_J': 0,
                                        'TRIG_ENGINE_OP_K': 1,
                                        'TRIG_ENGINE_OP_J_OR_K': 2,
                                        'TRIG_ENGINE_OP_J_AND_K': 3,
                                        'TRIG_ENGINE_OP_J_XOR_K': 4,
                                        'TRIG_ENGINE_OP_J_AND_NOT_K': 5,
                                        'TRIG_ENGINE_OP_NOT_J_AND_K': 6})
        n_trigger_engines = 2

        for i in range(1, n_trigger_engines+1):
            self.add_parameter(name=f'trigger_engine{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Engine {i}',
                               unit=None,
                               initial_value='TRIG_ENGINE_' + ('J' if i == 1 else 'K'),
                               val_mapping={'TRIG_ENGINE_J': 0,
                                            'TRIG_ENGINE_K': 1})
            self.add_parameter(name=f'trigger_source{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Source {i}',
                               unit=None,
                               initial_value='EXTERNAL',
                               val_mapping={'CHANNEL_A': 0,
                                            'CHANNEL_B': 1,
                                            'EXTERNAL': 2,
                                            'DISABLE': 3,
                                            'CHANNEL_C': 4,
                                            'CHANNEL_D': 5})
            self.add_parameter(name=f'trigger_slope{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Slope {i}',
                               unit=None,
                               initial_value='TRIG_SLOPE_POSITIVE',
                               val_mapping={'TRIG_SLOPE_POSITIVE': 1,
                                            'TRIG_SLOPE_NEGATIVE': 2})
            self.add_parameter(name=f'trigger_level{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Level {i}',
                               unit=None,
                               initial_value=140,
                               vals=validators.Ints(0, 255))
        self.add_parameter(name='external_trigger_coupling',
                           parameter_class=TraceParameter,
                           label='External Trigger Coupling',
                           unit=None,
                           initial_value='DC',
                           val_mapping={'AC': 1, 'DC': 2})
        self.add_parameter(name='external_trigger_range',
                           parameter_class=TraceParameter,
                           label='External Trigger Range',
                           unit=None,
                           initial_value='ETR_5V',
                           val_mapping={'ETR_5V': 0, 'ETR_TTL': 2})
        self.add_parameter(name='trigger_delay',
                           parameter_class=TraceParameter,
                           label='Trigger Delay',
                           unit='Sample clock cycles',
                           initial_value=0,
                           vals=validators.Multiples(divisor=8, min_value=0))
        self.add_parameter(name='timeout_ticks',
                           parameter_class=TraceParameter,
                           label='Timeout Ticks',
                           unit='10 us',
                           initial_value=0,
                           vals=validators.Ints(min_value=0))
        #  The card has two AUX I/O ports, which only AUX 2 is controlled by
        #  the software (AUX 1 is controlled by the firmware). The user should
        #  use AUX 2 for controlling the AUX via aux_io_mode and aux_io_param.
        self.add_parameter(name='aux_io_mode',
                           parameter_class=TraceParameter,
                           label='AUX I/O Mode',
                           unit=None,
                           initial_value='AUX_OUT_TRIGGER',
                           val_mapping={'AUX_OUT_TRIGGER': 0,
                                        'AUX_IN_TRIGGER_ENABLE': 1,
                                        'AUX_IN_AUXILIARY': 13})
        self.add_parameter(name='aux_io_param',
                           parameter_class=TraceParameter,
                           label='AUX I/O Param',
                           unit=None,
                           initial_value='NONE',
                           val_mapping={'NONE': 0,
                                        'TRIG_SLOPE_POSITIVE': 1,
                                        'TRIG_SLOPE_NEGATIVE': 2})

        #  The above parameters are important for preparing the card.
        self.add_parameter(name='mode',
                           label='Acquisition mode',
                           unit=None,
                           initial_value='NPT',
                           set_cmd=None,
                           val_mapping={'NPT': 0x200,
                                        'TS': 0x400})
        self.add_parameter(name='samples_per_record',
                           label='Samples per Record',
                           unit=None,
                           initial_value=1024,
                           set_cmd=None,
                           vals=validators.Multiples(
                                divisor=self.samples_divisor, min_value=256))
        self.add_parameter(name='records_per_buffer',
                           label='Records per Buffer',
                           unit=None,
                           initial_value=10,
                           set_cmd=None,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='buffers_per_acquisition',
                           label='Buffers per Acquisition',
                           unit=None,
                           set_cmd=None,
                           initial_value=10,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='channel_selection',
                           label='Channel Selection',
                           unit=None,
                           set_cmd=None,
                           initial_value='AB',
                           val_mapping={'A': 1,
                                        'B': 2,
                                        'AB': 3,
                                        'C': 4,
                                        'AC': 5,
                                        'BC': 6,
                                        'D': 7,
                                        'AD': 8,
                                        'BD': 9,
                                        'CD': 10,
                                        'ABCD': 11})
        self.add_parameter(name='transfer_offset',
                           label='Transfer Offset',
                           unit='Samples',
                           set_cmd=None,
                           initial_value=0,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='external_startcapture',
                           label='External Startcapture',
                           unit=None,
                           set_cmd=None,
                           initial_value='ENABLED',
                           val_mapping={'DISABLED': 0X0,
                                        'ENABLED': 0x1})
        self.add_parameter(name='enable_record_headers',
                           label='Enable Record Headers',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x8})
        self.add_parameter(name='alloc_buffers',
                           label='Alloc Buffers',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x20})
        self.add_parameter(name='fifo_only_streaming',
                           label='Fifo Only Streaming',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x800})
        self.add_parameter(name='interleave_samples',
                           label='Interleave Samples',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x1000})
        self.add_parameter(name='get_processed_data',
                           label='Get Processed Data',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x2000})
        self.add_parameter(name='allocated_buffers',
                           label='Allocated Buffers',
                           unit=None,
                           set_cmd=None,
                           initial_value=4,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='buffer_timeout',
                           label='Buffer Timeout',
                           unit='ms',
                           set_cmd=None,
                           initial_value=1000,
                           vals=validators.Ints(min_value=0))



import re

import numpy as np
import pytest
from hypothesis import given
from hypothesis.extra.numpy import complex_number_dtypes

from qcodes.utils.types import (concrete_complex_types, numpy_concrete_floats,
                                numpy_concrete_ints,
                                numpy_non_concrete_floats_instantiable,
                                numpy_non_concrete_ints_instantiable)
from qcodes.utils.validators import Arrays


def test_type():
    m = Arrays(min_value=0.0, max_value=3.2, shape=(2, 2))
    for v in ['somestring', 4, 2, [[2, 0], [1, 2]]]:
        with pytest.raises(TypeError):
            m.validate(v)


def test_complex_min_max_raises():
    """
    Min max is not implemented for complex types
    """
    with pytest.raises(TypeError, match=r"min_value must be a real number\."
                                        r" It is \(1\+1j\) of type "
                                        r"<class 'complex'>"):
        Arrays(min_value=1 + 1j)
    with pytest.raises(TypeError, match=r"max_value must be a real number. "
                                        r"It is \(1\+1j\) of type "
                                        r"<class 'complex'>"):
        Arrays(max_value=1 + 1j)
    with pytest.raises(TypeError, match=r'Setting min_value or max_value is '
                                        r'not supported for complex '
                                        r'validators'):
        Arrays(max_value=1, valid_types=(np.complexfloating,))


@given(dtype=complex_number_dtypes())
def test_complex(dtype):
    a = Arrays(valid_types=(np.complexfloating,))
    a.validate(np.arange(10, dtype=dtype))


def test_complex_subtypes():
    """Test that specifying a specific complex subtype works as expected"""
    a = Arrays(valid_types=(np.complex64,))

    a.validate(np.arange(10, dtype=np.complex64))
    with pytest.raises(TypeError, match=r"is not any of "
                                        r"\(<class 'numpy.complex64'>,\)"
                                        r" it is complex128"):
        a.validate(np.arange(10, dtype=np.complex128))
    a = Arrays(valid_types=(np.complex128,))

    a.validate(np.arange(10, dtype=np.complex128))
    with pytest.raises(TypeError, match=r"is not any of "
                                        r"\(<class 'numpy.complex128'>,\)"
                                        r" it is complex64"):
        a.validate(np.arange(10, dtype=np.complex64))


def test_min_max_real_ints_raises():
    with pytest.raises(TypeError, match="min_value must be an instance "
                                        "of valid_types."):
        Arrays(valid_types=(np.integer,), min_value=1.0)
    with pytest.raises(TypeError, match="max_value must be an instance "
                                        "of valid_types."):
        Arrays(valid_types=(np.integer,), max_value=6.0)


def test_min_max_ints_real_raises():
    with pytest.raises(TypeError, match="min_value must be an instance "
                                        "of valid_types."):
        Arrays(valid_types=(np.floating,), min_value=1)
    with pytest.raises(TypeError, match="max_value must be an instance "
                                        "of valid_types."):
        Arrays(valid_types=(np.floating,), max_value=6)


def test_real_subtypes():
    """
    Test that validating a concrete real type into an array that
    only support other concrete types raises as expected
    """
    types = list(numpy_concrete_ints + numpy_concrete_floats)
    randint = np.random.randint(0, len(types))
    mytype = types.pop(randint)

    a = Arrays(valid_types=(mytype,))
    a.validate(np.arange(10, dtype=mytype))
    a = Arrays(valid_types=types)
    with pytest.raises(TypeError, match=r'is not any of'):
        a.validate(np.arange(10, dtype=mytype))


def test_complex_default_raises():
    """Complex types should not validate by default"""
    a = Arrays()
    for dtype in concrete_complex_types:
        with pytest.raises(TypeError, match=r"is not any of \(<class "
                                            r"'numpy.integer'>, <class "
                                            r"'numpy.floating'>\) "
                                            r"it is complex"):
            a.validate(np.arange(10, dtype=dtype))


def test_text_type_raises():
    """Text types are not supported """
    with pytest.raises(TypeError, match="Arrays validator only supports "
                                        "numeric types: <class "
                                        "'numpy.str_'> is not supported."):
        Arrays(valid_types=(np.dtype('<U5').type,))


def test_text_array_raises():
    """Test that an array of text raises"""
    a = Arrays()
    with pytest.raises(TypeError,
                       match=r"type of \['A' 'BC' 'CDF'\] is not any of "
                             r"\(<class 'numpy.integer'>, <class "
                             r"'numpy.floating'>\) it is <U3;"):
        a.validate(np.array(['A', 'BC', 'CDF']))


def test_default_types():
    """Arrays constructed with all concrete and abstract real number
    types should validate by default"""
    a = Arrays()

    integer_types = (
            (int,)
            + numpy_non_concrete_ints_instantiable
            + numpy_concrete_ints
    )
    for mytype in integer_types:
        a.validate(np.arange(10, dtype=mytype))

    float_types = (
            (float,)
            + numpy_non_concrete_floats_instantiable
            + numpy_concrete_floats)
    for mytype in float_types:
        a.validate(np.arange(10, dtype=mytype))


def test_min_max():
    m = Arrays(min_value=-5, max_value=50, shape=(2, 2))
    v = np.array([[2, 0], [1, 2]])
    m.validate(v)
    v = 100 * v
    with pytest.raises(ValueError):
        m.validate(v)
    v = -1 * v
    with pytest.raises(ValueError):
        m.validate(v)

    m = Arrays(min_value=-5, shape=(2, 2))
    v = np.array([[2, 0], [1, 2]])
    m.validate(v * 100)


def test_max_smaller_min_raises():
    with pytest.raises(TypeError, match='max_value must be '
                                        'bigger than min_value'):
        Arrays(min_value=10, max_value=-10)


def test_shape():
    m = Arrays(min_value=-5, max_value=50, shape=(2, 2))

    v1 = np.array([[2, 0], [1, 2]])
    v2 = np.array([[2, 0], [1, 2], [2, 3]])

    # v1 is the correct shape but v2 is not
    m.validate(v1)
    with pytest.raises(ValueError):
        m.validate(v2)
    # both should pass if no shape specified
    m = Arrays(min_value=-5, max_value=50)
    m.validate(v1)
    m.validate(v2)


def test_shape_defered():
    m = Arrays(min_value=-5, max_value=50, shape=(lambda: 2, lambda: 2))
    v1 = np.array([[2, 5], [3, 7]])
    m.validate(v1)
    v2 = np.array([[2, 0], [1, 2], [2, 3]])
    with pytest.raises(ValueError):
        m.validate(v2)


def test_valid_values_with_shape():
    val = Arrays(min_value=-5, max_value=50, shape=(2, 2))
    for vval in val.valid_values:
        val.validate(vval)


def test_valid_values():
    val = Arrays(min_value=-5, max_value=50)
    for vval in val.valid_values:
        val.validate(vval)


def test_shape_non_sequence_raises():
    with pytest.raises(ValueError):
        _ = Arrays(shape=5)
    with pytest.raises(ValueError):
        _ = Arrays(shape=lambda: 10)


def test_repr():
    a = Arrays()
    assert str(a) == '<Arrays, shape: None>'
    b = Arrays(min_value=1, max_value=2)
    assert str(b) == '<Arrays 1<=v<=2, shape: None>'
    c = Arrays(shape=(2, 2))
    assert str(c) == '<Arrays, shape: (2, 2)>'
    c = Arrays(shape=(lambda: 2, 2))
    assert re.match(r"<Arrays, shape: \(<function "
                    r"test_repr.<locals>.<lambda> "
                    r"at 0x[a-fA-f0-9]*>, 2\)>", str(c))



from typing import Optional, TYPE_CHECKING, List, Any

from .KeysightB1517A import B1517A
from .constants import IMeasRange, IOutputRange

if TYPE_CHECKING:
    from .KeysightB1500_base import KeysightB1500


class B1511B(B1517A):
    """
    Driver for Keysight B1511B Source/Monitor Unit module for B1500
    Semiconductor Parameter Analyzer.

    Args:
        parent: mainframe B1500 instance that this module belongs to
        name: Name of the instrument instance to create. If `None`
            (Default), then the name is autogenerated from the instrument
            class.
        slot_nr: Slot number of this module (not channel number)
        asu_present: Flag to acknowledge ASU presence
    """

    def __init__(self, parent: 'KeysightB1500', name: Optional[str],
                 slot_nr: int, **kwargs: Any):
        super().__init__(parent, name, slot_nr, **kwargs)
        self._valid_i_measure_ranges: List[IMeasRange] = [IMeasRange.AUTO,
                                                          IMeasRange.MIN_1nA,
                                                          IMeasRange.MIN_10nA,
                                                          IMeasRange.MIN_100nA,
                                                          IMeasRange.MIN_1uA,
                                                          IMeasRange.MIN_10uA,
                                                          IMeasRange.MIN_100uA,
                                                          IMeasRange.MIN_1mA,
                                                          IMeasRange.MIN_10mA,
                                                          IMeasRange.MIN_100mA,
                                                          IMeasRange.FIX_1nA,
                                                          IMeasRange.FIX_10nA,
                                                          IMeasRange.FIX_100nA,
                                                          IMeasRange.FIX_1uA,
                                                          IMeasRange.FIX_10uA,
                                                          IMeasRange.FIX_100uA,
                                                          IMeasRange.FIX_1mA,
                                                          IMeasRange.FIX_10mA,
                                                          IMeasRange.FIX_100mA]
        self._asu_valid_i_measure_ranges: List[IMeasRange] = [
            IMeasRange.MIN_1pA, IMeasRange.MIN_10pA, IMeasRange.MIN_100pA,
            IMeasRange.FIX_1pA, IMeasRange.FIX_10pA, IMeasRange.FIX_100pA]
        self._valid_i_output_ranges: List[IOutputRange] = [
            IOutputRange.AUTO, IOutputRange.MIN_1nA, IOutputRange.MIN_10nA,
            IOutputRange.MIN_100nA, IOutputRange.MIN_1uA,
            IOutputRange.MIN_10uA, IOutputRange.MIN_100uA,
            IOutputRange.MIN_1mA, IOutputRange.MIN_10mA, IOutputRange.MIN_100mA]
        self._asu_valid_i_output_ranges: List[IOutputRange] = [
            IOutputRange.MIN_1pA, IOutputRange.MIN_10pA, IOutputRange.MIN_100pA]

        self.asu_present: bool = False

    @property
    def asu_present(self) -> bool:
        return self._asu_present

    @asu_present.setter
    def asu_present(self, val: bool) -> None:
        if not isinstance(val, bool):
            raise TypeError("Expected: True or False")

        self._asu_present = val

        if self.asu_present:
            self._valid_i_measure_ranges = self._valid_i_measure_ranges + \
                                           self._asu_valid_i_measure_ranges
            self._valid_i_output_ranges = self._valid_i_output_ranges + \
                                          self._asu_valid_i_output_ranges
        else:
            self._valid_i_measure_ranges = list(
                set(self._valid_i_measure_ranges) -
                set(self._asu_valid_i_measure_ranges)
            )
            self._valid_i_output_ranges = list(
                set(self._valid_i_output_ranges) -
                set(self._asu_valid_i_output_ranges)
            )



from typing import Any, List, Sequence, Tuple

from qcodes import ChannelList, InstrumentChannel, VisaInstrument
from qcodes import validators as vals


class RigolDP8xxChannel(InstrumentChannel):
    def __init__(
            self,
            parent: "_RigolDP8xx",
            name: str,
            channel: int,
            ch_range: Tuple[float, float],
            ovp_range: Tuple[float, float],
            ocp_range: Tuple[float, float]
    ):
        super().__init__(parent, name)

        self.vmax = ch_range[0]
        self.imax = ch_range[1]
        self.ovp_range = ovp_range
        self.ocp_range = ocp_range

        select_cmd = f":INSTrument:NSELect {channel};"
        strstrip = lambda s: str(s).strip()

        self.add_parameter("set_voltage",
                           label='Target voltage output',
                           set_cmd="{} :SOURce:VOLTage:LEVel:IMMediate:AMPLitude {}".format(
                               select_cmd, '{}'),
                           get_cmd="{} :SOURce:VOLTage:LEVel:IMMediate:AMPLitude?".format(
                               select_cmd),
                           get_parser=float,
                           unit='V',
                           vals=vals.Numbers(min(0, self.vmax), max(0, self.vmax))
                          )
        self.add_parameter("set_current",
                           label='Target current output',
                           set_cmd="{} :SOURce:CURRent:LEVel:IMMediate:AMPLitude {}".format(
                               select_cmd, '{}'),
                           get_cmd="{} :SOURce:CURRent:LEVel:IMMediate:AMPLitude?".format(
                               select_cmd),
                           get_parser=float,
                           unit='A',
                           vals=vals.Numbers(0, self.imax)
                           )
        self.add_parameter('state',
                           label='Output enabled',
                           set_cmd='{} :OUTPut:STATe {}'.format(select_cmd, '{}'),
                           get_cmd=f'{select_cmd} :OUTPut:STATe?',
                           get_parser=strstrip,
                           vals=vals.OnOff()
                           )
        self.add_parameter('mode',
                           label='Get the output mode',
                           get_cmd=f'{select_cmd} :OUTPut:MODE?',
                           get_parser=strstrip,
                           val_mapping={'ConstantVoltage': 'CV',
                                        'ConstantCurrent': 'CC',
                                        'Unregulated': 'UR'}
                          )
        self.add_parameter("voltage",
                           label='Measured voltage',
                           get_cmd="{} :MEASure:VOLTage:DC?".format(
                               select_cmd),
                           get_parser=float,
                           unit='V',
                          )
        self.add_parameter("current",
                           label='Measured current',
                           get_cmd="{} :MEASure:CURRent:DC?".format(
                               select_cmd),
                           get_parser=float,
                           unit='A',
                           )
        self.add_parameter("power",
                           label='Measured power',
                           get_cmd="{} :MEASure:POWer?".format(
                               select_cmd),
                           get_parser=float,
                           unit='W',
                           )
        self.add_parameter("ovp_value",
                           label='Over Voltage Protection value',
                           set_cmd="{} :VOLTage:PROTection:LEVel {}".format(
                               select_cmd, '{}'),
                           get_cmd="{} :VOLTage:PROTection:LEVel?".format(
                               select_cmd),
                           get_parser=float,
                           unit='V',
                           vals=vals.Numbers(self.ovp_range[0], self.ovp_range[1])
                           )
        self.add_parameter('ovp_state',
                           label='Over Voltage Protection status',
                           set_cmd='{} :VOLTage:PROTection:STATe {}'.format(select_cmd, '{}'),
                           get_cmd=f'{select_cmd} :VOLTage:PROTection:STATe?',
                           get_parser=strstrip,
                           vals=vals.OnOff()
                           )
        self.add_parameter("ocp_value",
                           label='Over Current Protection value',
                           set_cmd="{} :CURRent:PROTection:LEVel {}".format(
                               select_cmd, '{}'),
                           get_cmd="{} :CURRent:PROTection:LEVel?".format(
                               select_cmd),
                           get_parser=float,
                           unit='A',
                           vals=vals.Numbers(self.ocp_range[0], self.ocp_range[1])
                           )
        self.add_parameter('ocp_state',
                           label='Over Current Protection status',
                           set_cmd='{} :CURRent:PROTection:STATe {}'.format(select_cmd, '{}'),
                           get_cmd=f'{select_cmd} :CURRent:PROTection:STATe?',
                           get_parser=strstrip,
                           vals=vals.OnOff()
                           )


class _RigolDP8xx(VisaInstrument):
    """
    This is the general DP8xx Power Supply driver class that implements shared parameters and functionality
    among all similar power supply from Rigole.

    This driver was written to be inherited from by a specific driver (e.g. DP832).
    """

    def __init__(
            self,
            name: str,
            address: str,
            channels_ranges: Sequence[Tuple[float, float]],
            ovp_ranges: Tuple[Sequence[Tuple[float, float]],
                              Sequence[Tuple[float, float]]],
            ocp_ranges: Tuple[Sequence[Tuple[float, float]],
                              Sequence[Tuple[float, float]]],
            **kwargs: Any
    ):
        super().__init__(name, address, **kwargs)

        # Check if precision extension has been installed
        opt = self.installed_options()
        if 'DP8-ACCURACY' in opt:
            ovp_ranges_selected = ovp_ranges[1]
            ocp_ranges_selected = ocp_ranges[1]
        else:
            ovp_ranges_selected = ovp_ranges[0]
            ocp_ranges_selected = ocp_ranges[0]

        # channel-specific parameters
        channels = ChannelList(self, "SupplyChannel", RigolDP8xxChannel, snapshotable=False)
        for ch_num, channel_range in enumerate(channels_ranges):
            ch_name = f"ch{ch_num + 1}"
            channel = RigolDP8xxChannel(
                self,
                ch_name,
                ch_num + 1,
                channel_range,
                ovp_ranges_selected[ch_num],
                ocp_ranges_selected[ch_num]
            )
            channels.append(channel)
            self.add_submodule(ch_name, channel)
        channels.lock()
        self.add_submodule("channels", channels)

        self.connect_message()

    def installed_options(self) -> List[str]:
        """Return the installed options"""

        opt = self.ask("*OPT?")
        optl = opt.strip().split(',')
        optl_clean = [x for x in optl if x != '0']
        return optl_clean



import numpy as np
from typing import cast, Optional, List, Union, Sequence, Any, Tuple, Dict, Type, Set
from types import TracebackType

from qcodes import VisaInstrument, InstrumentChannel
from qcodes.instrument.parameter import invert_val_mapping, Parameter, \
    DelegateParameter, MultiParameter, ParamRawDataType
from qcodes.utils.validators import Enum, Numbers, Ints, Lists, Arrays
from qcodes.utils.helpers import create_on_off_val_mapping


class DataArray7510(MultiParameter):
    """
    Data class when user selected more than one field for data output.
    """
    _data: Tuple[Tuple[Any, ...], ...] = ((), ())

    def __init__(self,
                 names: Sequence[str],
                 shapes: Sequence[Sequence[int]],
                 setpoints: Optional[Sequence[Sequence[Any]]],
                 **kwargs: Any):
        super().__init__(name='data_array_7510',
                         names=names,
                         shapes=shapes,
                         setpoints=setpoints,
                         **kwargs)
        for param_name in self.names:
            self.__dict__.update({param_name: []})

    def get_raw(self) -> Optional[Tuple[ParamRawDataType, ...]]:
        return self._data


class GeneratedSetPoints(Parameter):
    """
    A parameter that generates a setpoint array from start, stop and num points
    parameters.
    """
    def __init__(self,
                 start: Parameter,
                 stop: Parameter,
                 n_points: Parameter,
                 *args: Any,
                 **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._start = start
        self._stop = stop
        self._n_points = n_points

    def get_raw(self) -> np.ndarray:
        start = self._start()
        assert start is not None
        stop = self._stop()
        assert stop is not None
        n_points = self._n_points()
        assert n_points is not None

        return np.linspace(start, stop, n_points)


class Buffer7510(InstrumentChannel):
    """
    Treat the reading buffer as a submodule, similar to Sense.
    """
    default_buffer = {"defbuffer1", "defbuffer2"}

    buffer_elements = {
        "date": "DATE",
        "measurement_formatted": "FORMatted",
        "fractional_seconds": "FRACtional",
        "measurement": "READing",
        "relative_time": "RELative",
        "seconds": "SEConds",
        "measurement_status": "STATus",
        "time": "TIME",
        "timestamp": "TSTamp",
        "measurement_unit": "UNIT"
    }

    inverted_buffer_elements = invert_val_mapping(buffer_elements)

    def __init__(
            self,
            parent: 'Keithley7510',
            name: str,
            size: Optional[int] = None,
            style: str = ''
    ) -> None:
        super().__init__(parent, name)
        self._size = size
        self.style = style

        if self.short_name not in self.default_buffer:
            # when making a new buffer, the "size" parameter is required.
            if size is None:
                raise TypeError(
                    "buffer() missing 1 required positional argument: 'size'"
                )
            self.write(
                f":TRACe:MAKE '{self.short_name}', {self._size}, {self.style}"
            )
        else:
            # when referring to default buffer, "size" parameter is not needed.
            if size is not None:
                self.log.warning(
                    f"Please use method 'size()' to resize default buffer "
                    f"{self.short_name} size to {self._size}."
                )

        self.add_parameter(
            "size",
            get_cmd=f":TRACe:POINts? '{self.short_name}'",
            set_cmd=f":TRACe:POINts {{}}, '{self.short_name}'",
            get_parser=int,
            docstring="The number of readings a buffer can store."
        )

        self.add_parameter(
            "number_of_readings",
            get_cmd=f":TRACe:ACTual? '{self.short_name}'",
            get_parser=int,
            docstring="Get the number of readings in the reading buffer."
        )

        self.add_parameter(
            "last_index",
            get_cmd=f":TRACe:ACTual:END? '{self.short_name}'",
            get_parser=int,
            docstring="Get the last index of readings in the reading buffer."
        )

        self.add_parameter(
            "first_index",
            get_cmd=f":TRACe:ACTual:STARt? '{self.short_name}'",
            get_parser=int,
            docstring="Get the starting index of readings in the reading "
                      "buffer."
        )

        self.add_parameter(
            "data_start",
            initial_value=1,
            get_cmd=None,
            set_cmd=None,
            docstring="First index of the data to be returned."
        )

        self.add_parameter(
            "data_end",
            initial_value=1,
            get_cmd=None,
            set_cmd=None,
            docstring="Last index of the data to be returned."
        )

        self.add_parameter(
            "elements",
            get_cmd=None,
            get_parser=self._from_scpi_to_name,
            set_cmd=None,
            set_parser=self._from_name_to_scpi,
            vals=Lists(Enum(*list(self.buffer_elements.keys()))),
            docstring="List of buffer elements to read."
        )

        self.add_parameter(
            "setpoints_start",
            label="start value for the setpoints",
            source=None,
            parameter_class=DelegateParameter
        )

        self.add_parameter(
            "setpoints_stop",
            label="stop value for the setpoints",
            source=None,
            parameter_class=DelegateParameter
        )

        self.add_parameter(
            "n_pts",
            label="total n for the setpoints",
            get_cmd=self._get_n_pts
        )

        self.add_parameter(
            "setpoints",
            parameter_class=GeneratedSetPoints,
            start=self.setpoints_start,
            stop=self.setpoints_stop,
            n_points=self.n_pts,
            vals=Arrays(shape=(self.n_pts.get_latest,))
        )

        self.add_parameter(
            "t_start",
            label="start time",
            unit="s",
            initial_value=0,
            get_cmd=None,
            set_cmd=None,
            set_parser=float
        )

        self.add_parameter(
            "t_stop",
            label="stop time",
            unit="s",
            initial_value=1,
            get_cmd=None,
            set_cmd=None,
            set_parser=float
        )

        self.add_parameter(
            "fill_mode",
            get_cmd=f":TRACe:FILL:MODE? '{self.short_name}'",
            set_cmd=f":TRACe:FILL:MODE {{}}, '{self.short_name}'",
            vals=Enum('CONT', 'continuous', 'ONCE', 'once'),
            docstring="if a reading buffer is filled continuously or is filled"
                      " once and stops"
        )

    def _from_name_to_scpi(self, element_names: List[str]) -> List[str]:
        return [self.buffer_elements[element] for element in element_names]

    def _from_scpi_to_name(self, element_scpis: List[str]) -> List[str]:
        if element_scpis is None:
            return []
        return [
            self.inverted_buffer_elements[element] for element in element_scpis
        ]

    def _get_n_pts(self) -> int:
        return self.data_end() - self.data_start() + 1

    def set_setpoints(self,
                      start: Parameter,
                      stop: Parameter,
                      label: Optional[str] = None) -> None:
        self.setpoints_start.source = start
        self.setpoints_stop.source = stop
        self.setpoints.unit = start.unit
        if label is not None:
            self.setpoints.label = label

    def __enter__(self) -> "Buffer7510":
        return self

    def __exit__(self, exception_type: Optional[Type[BaseException]],
                 value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        self.delete()

    @property
    def available_elements(self) -> Set[str]:
        return set(self.buffer_elements.keys())

    @property
    def n_elements(self) -> int:
        return max(1, len(self.elements()))

    @property
    def data(self) -> DataArray7510:
        return self._get_data()

    def get_last_reading(self) -> str:
        """
        This method requests the latest reading from a reading buffer.

        """
        if not self.elements():
            return self.ask(f":FETCh? '{self.short_name}'")
        fetch_elements = [
            self.buffer_elements[element] for element in self.elements()
        ]
        return self.ask(
            f":FETCh? '{self.short_name}', {','.join(fetch_elements)}"
        )

    def _get_data(self) -> DataArray7510:
        """
        This command returns the data in the buffer, depends on the user
        selected elements.
        """
        try:
            _ = self.setpoints()
        except NotImplementedError:
            # if the "setpionts" has not been implemented, use a time series
            # with parameters "t_start" and "t_stop":
            self.set_setpoints(self.t_start, self.t_stop)

        if self.parent.digi_sense_function() == "None":
            # when current sense is not digitize sense
            sense_function = self.parent.sense_function()
            unit = Sense7510.function_modes[sense_function]["unit"]
        else:
            # when current sense is digitize sense
            sense_function = self.parent.digi_sense_function()
            unit = DigitizeSense7510.function_modes[sense_function]["unit"]

        elements_units = {
            "date": "str",
            "measurement_formatted": "str",
            "fractional_seconds": "s",
            "measurement": unit,
            "relative_time": "s",
            "seconds": "s",
            "measurement_status": "",
            "time": "str",
            "timestamp": "str",
            "measurement_unit": "str"
        }

        if not self.elements():
            raw_data = self.ask(f":TRACe:DATA? "
                                f"{self.data_start()}, "
                                f"{self.data_end()}, "
                                f"'{self.short_name}'")
        else:
            elements = \
                [self.buffer_elements[element] for element in self.elements()]
            raw_data = self.ask(f":TRACe:DATA? {self.data_start()}, "
                                f"{self.data_end()}, "
                                f"'{self.short_name}', "
                                f"{','.join(elements)}")

        all_data = raw_data.split(",")

        if len(self.elements()) == 0:
            elements = ['measurement']
        else:
            elements = self.elements()
        n_elements = len(elements)

        units = tuple(elements_units[element] for element in elements)
        processed_data = dict.fromkeys(elements)
        for i, (element, unit) in enumerate(zip(elements, units)):
            if unit == 'str':
                processed_data[element] = np.array(all_data[i::n_elements])
            else:
                processed_data[element] = np.array([
                    float(v) for v in all_data[i::n_elements]
                ])

        data = DataArray7510(
            names=tuple(elements),
            shapes=((self.n_pts(),),) * n_elements,
            units=units,
            setpoints=((self.setpoints(),),) * n_elements,
            setpoint_units=((self.setpoints.unit,),) * n_elements,
            setpoint_names=((self.setpoints.label,),) * n_elements
        )
        data._data = tuple(
            tuple(processed_data[element]) for element in elements
        )
        for i in range(len(data.names)):
            setattr(data, data.names[i], tuple(processed_data[data.names[i]]))
        return data

    def clear_buffer(self) -> None:
        """
        Clear the data in the buffer
        """
        self.write(f":TRACe:CLEar '{self.short_name}'")

    def trigger_start(self) -> None:
        """
        This method makes readings using the active measure function and
        stores them in a reading buffer.
        """
        self.write(f":TRACe:TRIGger '{self.short_name}'")

    def delete(self) -> None:
        if self.short_name not in self.default_buffer:
            self.parent.submodules.pop(f"_buffer_{self.short_name}")
            self.parent.buffer_name("defbuffer1")
            self.write(f":TRACe:DELete '{self.short_name}'")


class Sense7510(InstrumentChannel):
    """
    The sense module of the Keithley 7510 DMM, based on the sense module of
    Keithley 2450 SMU.

    Args:
        parent
        name
        proper_function: This can be one of modes listed in the dictionary
            "function_modes", e.g.,  "current", "voltage", or "resistance".
            "voltage"/"current" is for DC voltage/current.
            "Avoltage"/"Acurrent" is for AC voltage/current.
            "resistance" is for two-wire measurement of resistance.
            "Fresistance" is for Four-wire measurement of resistance.

            All parameters and methods in this submodule should only be
            accessible to the user if
            self.parent.sense_function.get() == self._proper_function. We
            ensure this through the 'sense' property on the main driver class
            which returns the proper submodule for any given function mode.
    """

    function_modes = {
        "voltage": {
            "name": '"VOLT:DC"',
            "unit": 'V',
            "range_vals": Numbers(0.1, 1000),
        },
        "Avoltage": {
            "name": '"VOLT:AC"',
            "unit": 'V',
            "range_vals": Numbers(0.1, 700),
        },
        "current": {
            "name": '"CURR:DC"',
            "unit": 'V',
            "range_vals": Numbers(10e-6, 10),
        },
        "Acurrent": {
            "name": '"CURR:AC"',
            "unit": 'A',
            "range_vals": Numbers(1e-3, 10),
        },
        "resistance": {
            "name": '"RES"',
            "unit": 'Ohm',
            "range_vals": Numbers(10, 1e9),
        },
        "Fresistance": {
            "name": '"FRES"',
            "unit": 'Ohm',
            "range_vals": Numbers(1, 1e9),
        }
    }

    def __init__(
            self,
            parent: VisaInstrument,
            name: str,
            proper_function: str
    ) -> None:

        super().__init__(parent, name)

        self._proper_function = proper_function
        range_vals = self.function_modes[self._proper_function]["range_vals"]
        unit = self.function_modes[self._proper_function]["unit"]

        self.function = self.parent.sense_function

        self.add_parameter(
            self._proper_function,
            get_cmd=self._measure,
            get_parser=float,
            unit=unit,
            docstring="Make measurements, place them in a reading buffer, and "
                      "return the last reading."
        )

        self.add_parameter(
            "auto_range",
            get_cmd=f":SENSe:{self._proper_function}:RANGe:AUTO?",
            set_cmd=f":SENSe:{self._proper_function}:RANGe:AUTO {{}}",
            val_mapping=create_on_off_val_mapping(on_val="1", off_val="0"),
            docstring="Determine if the measurement range is set manually or "
                      "automatically for the selected measure function."
        )

        self.add_parameter(
            "range",
            get_cmd=f":SENSe:{self._proper_function}:RANGe?",
            set_cmd=f":SENSe:{self._proper_function}:RANGe {{}}",
            vals=range_vals,
            get_parser=float,
            unit=unit,
            docstring="Determine the positive full-scale measure range."
        )

        self.add_parameter(
            "nplc",
            get_cmd=f":SENSe:{self._proper_function}:NPLCycles?",
            set_cmd=f":SENSe:{self._proper_function}:NPLCycles {{}}",
            vals=Numbers(0.01, 10),
            get_parser=float,
            docstring="Set the time that the input signal is measured for the "
                      "selected function.(NPLC = number of power line cycles)"
        )

        self.add_parameter(
            "auto_delay",
            get_cmd=f":SENSe:{self._proper_function}:DELay:AUTO?",
            set_cmd=f":SENSe:{self._proper_function}:DELay:AUTO {{}}",
            val_mapping=create_on_off_val_mapping(on_val="ON", off_val="OFF"),
            docstring="Enable or disable the automatic delay that occurs "
                      "before each measurement."
        )

        self.add_parameter(
            'user_number',
            get_cmd=None,
            set_cmd=None,
            vals=Ints(1, 5),
            docstring="Set the user number for user-defined delay."
        )

        self.add_parameter(
            "user_delay",
            get_cmd=self._get_user_delay,
            set_cmd=self._set_user_delay,
            vals=Numbers(0, 1e4),
            unit='second',
            docstring="Set a user-defined delay that you can use in the "
                      "trigger model."
        )

        self.add_parameter(
            "auto_zero",
            get_cmd=f":SENSe:{self._proper_function}:AZERo?",
            set_cmd=f":SENSe:{self._proper_function}:AZERo {{}}",
            val_mapping=create_on_off_val_mapping(on_val="1", off_val="0"),
            docstring="Enable or disable automatic updates to the internal "
                      "reference measurements (autozero) of the instrument."
        )

        self.add_parameter(
            "auto_zero_once",
            set_cmd=f":SENSe:AZERo:ONCE",
            docstring="Cause the instrument to refresh the reference and "
                      "zero measurements once"
        )

        self.add_parameter(
            "average",
            get_cmd=f":SENSe:{self._proper_function}:AVERage?",
            set_cmd=f":SENSe:{self._proper_function}:AVERage {{}}",
            val_mapping=create_on_off_val_mapping(on_val="1", off_val="0"),
            docstring="Enable or disable the averaging filter for measurements "
                      "of the selected function."
        )

        self.add_parameter(
            "average_count",
            get_cmd=f":SENSe:{self._proper_function}:AVERage:COUNt?",
            set_cmd=f":SENSe:{self._proper_function}:AVERage:COUNt {{}}",
            vals=Numbers(1, 100),
            docstring="Set the number of measurements that are averaged when "
                      "filtering is enabled."
        )

        self.add_parameter(
            "average_type",
            get_cmd=f":SENSe:{self._proper_function}:AVERage:TCONtrol?",
            set_cmd=f":SENSe:{self._proper_function}:AVERage:TCONtrol {{}}",
            vals=Enum('REP', 'rep', 'MOV', 'mov'),
            docstring="Set the type of averaging filter that is used for the "
                      "selected measure function when the measurement filter "
                      "is enabled."
        )

    def _get_user_delay(self) -> str:
        get_cmd = f":SENSe:{self._proper_function}:DELay:USER" \
                  f"{self.user_number()}?"
        return self.ask(get_cmd)

    def _set_user_delay(self, value: float) -> None:
        set_cmd = f":SENSe:{self._proper_function}:DELay:USER" \
                  f"{self.user_number()} {value}"
        self.write(set_cmd)

    def _measure(self) -> Union[float, str]:
        buffer_name = self.parent.buffer_name()
        return float(self.ask(f":MEASure? '{buffer_name}'"))

    def clear_trace(self, buffer_name: str = "defbuffer1") -> None:
        """
        Clear the data buffer
        """
        self.write(f":TRACe:CLEar '{buffer_name}'")


class DigitizeSense7510(InstrumentChannel):
    """
    The Digitize sense module of the Keithley 7510 DMM.
    """
    function_modes: Dict[str, Dict[str, Any]] = {
        "None": {
            "name": '"NONE"',
            "unit": '',
            "range_vals": None
        },
        "voltage": {
            "name": '"VOLT"',
            "unit": 'V',
            "range_vals": Numbers(0.1, 1000),
        },
        "current": {
            "name": '"CURR"',
            "unit": 'V',
            "range_vals": Numbers(10e-6, 10),
        }
    }

    def __init__(
            self,
            parent: VisaInstrument,
            name: str,
            proper_function: str
    ) -> None:

        super().__init__(parent, name)

        self._proper_function = proper_function
        range_vals = self.function_modes[self._proper_function]["range_vals"]
        unit = self.function_modes[self._proper_function]["unit"]

        self.function = self.parent.digi_sense_function

        self.add_parameter(
            self._proper_function,
            get_cmd=self._measure,
            unit=unit,
            docstring="Make measurements, place them in a reading buffer, and "
                      "return the last reading."
        )

        self.add_parameter(
            "range",
            get_cmd=f":SENSe:DIGitize:{self._proper_function}:RANGe?",
            set_cmd=f":SENSe:DIGitize:{self._proper_function}:RANGe {{}}",
            vals=range_vals,
            get_parser=float,
            unit=unit,
            docstring="Determine the positive full-scale measure range."
        )

        self.add_parameter(
            "input_impedance",
            get_cmd=":SENSe:DIGitize:VOLTage:INPutimpedance?",
            set_cmd=":SENSe:DIGitize:VOLTage:INPutimpedance {}",
            vals=Enum("AUTO", "MOHM10"),
            docstring="Determine when the 10 MΩ input divider is enabled. "
                      "'MOHM10' means 10 MΩ for all ranges."
        )

        self.add_parameter(
            'acq_rate',
            get_cmd=f":SENSe:DIGitize:{self._proper_function}:SRATE?",
            set_cmd=f":SENSe:DIGitize:{self._proper_function}:SRATE {{}}",
            vals=Ints(1000, 1000000),
            docstring="Define the precise acquisition rate at which the "
                      "digitizing measurements are made."
        )

        self.add_parameter(
            "aperture",
            get_cmd=f":SENSe:DIGitize:{self._proper_function}:APERture?",
            set_cmd=f":SENSe:DIGitize:{self._proper_function}:APERture {{}}",
            unit="us",
            docstring="Determine the aperture setting."
        )

        self.add_parameter(
            "count",
            get_cmd="SENSe:DIGitize:COUNt?",
            set_cmd="SENSe:DIGitize:COUNt {}",
            vals=Ints(1, 55000000),
            docstring="Set the number of measurements to digitize when a "
                      "measurement is requested"
        )

    def _measure(self) -> Union[float, str]:
        buffer_name = self.parent.buffer_name()
        return float(self.ask(f":MEASure:DIGitize? '{buffer_name}'"))


class Keithley7510(VisaInstrument):
    """
    The QCoDeS driver for the Keithley 7510 DMM
    """
    def __init__(self, name: str, address: str,
                 terminator: str = '\n', **kwargs: Any):
        """
        Create an instance of the instrument.

        Args:
            name: Name of the instrument instance
            address: Visa-resolvable instrument address
        """
        super().__init__(name, address, terminator=terminator, **kwargs)

        self.add_parameter(
            "sense_function",
            set_cmd=":SENSe:FUNCtion {}",
            get_cmd=":SENSe:FUNCtion?",
            val_mapping={
                key: value["name"]
                for key, value in Sense7510.function_modes.items()
            },
            docstring="Add sense functions listed in the function modes."
        )

        self.add_parameter(
            "digi_sense_function",
            set_cmd=":DIGitize:FUNCtion {}",
            get_cmd=":DIGitize:FUNCtion?",
            val_mapping={
                key: value["name"]
                for key, value in DigitizeSense7510.function_modes.items()
            },
            docstring="Make readings using the active digitize function."
        )

        self.add_parameter(
            "buffer_name",
            get_cmd=None,
            set_cmd=None,
            docstring="Name of the reading buffer in use."
        )

        self.add_parameter(
            "trigger_block_list",
            get_cmd=":TRIGger:BLOCk:LIST?",
            docstring="Return the settings for all trigger model blocks."
        )

        self.add_parameter(
            "trigger_in_ext_clear",
            set_cmd=":TRIGger:EXTernal:IN:CLEar",
            docstring="Clear the trigger event on the external in line."
        )

        self.add_parameter(
            "trigger_in_ext_edge",
            get_cmd=":TRIGger:EXTernal:IN:EDGE?",
            set_cmd=":TRIGger:EXTernal:IN:EDGE {}",
            vals=Enum("FALL", "RIS", "falling", "rising", "EITH", "either"),
            docstring="Type of edge that is detected as an input on the "
                      "external trigger in line"
        )

        self.add_parameter(
            "overrun_status",
            get_cmd=":TRIGger:EXTernal:IN:OVERrun?",
            docstring="Return the event detector overrun status."
        )

        self.add_parameter(
            "digitize_trigger",
            get_cmd=":TRIGger:DIGitize:STIMulus?",
            set_cmd=":TRIGger:DIGitize:STIMulus {}",
            vals=Enum("EXT", "external", "NONE"),
            docstring="Set the instrument to digitize a measurement the next "
                      "time it detects the specified trigger event."
        )

        self.add_parameter(
            "system_errors",
            get_cmd=":SYSTem:ERRor?",
            docstring="Return the oldest unread error message from the event "
                      "log and removes it from the log."
        )

        for proper_sense_function in Sense7510.function_modes:
            self.add_submodule(
                f"_sense_{proper_sense_function}",
                Sense7510(self, "sense", proper_sense_function)
            )

        for proper_sense_function in DigitizeSense7510.function_modes:
            self.add_submodule(
                f"_digi_sense_{proper_sense_function}",
                DigitizeSense7510(self, "digi_sense", proper_sense_function)
            )

        self.buffer_name('defbuffer1')
        self.buffer(name=self.buffer_name())
        self.connect_message()

    @property
    def sense(self) -> Sense7510:
        """
        We have different sense modules depending on the sense function.

        Return the correct source module based on the sense function.
        """
        sense_function = \
            self.sense_function.get_latest() or self.sense_function()
        submodule = self.submodules[f"_sense_{sense_function}"]
        return cast(Sense7510, submodule)

    @property
    def digi_sense(self) -> DigitizeSense7510:
        """
        We have different sense modules depending on the sense function.

        Return the correct source module based on the sense function.
        """
        if self.digi_sense_function() == 'None':
            raise AttributeError("Please use 'digi_sense_function()' to select"
                                 " a digitize function first")
        sense_function = \
            self.digi_sense_function.get_latest() or self.digi_sense_function()
        submodule = self.submodules[f"_digi_sense_{sense_function}"]
        return cast(DigitizeSense7510, submodule)

    def buffer(
            self,
            name: str,
            size: Optional[int] = None,
            style: str = ''
    ) -> Buffer7510:
        self.buffer_name(name)
        if f"_buffer_{name}" in self.submodules:
            return cast(Buffer7510, self.submodules[f"_buffer_{name}"])
        new_buffer = Buffer7510(parent=self, name=name, size=size, style=style)
        self.add_submodule(f"_buffer_{name}", new_buffer)
        return new_buffer

    def initiate(self) -> None:
        """
        This command starts the trigger model.
        """
        self.write(":INITiate")

    def wait(self) -> None:
        """
        This command postpones the execution of subsequent commands until all
        previous overlapped commands are finished.
        """
        self.write("*WAI")

    def clear_status(self) -> None:
        """
        This command clears the event registers of the Questionable Event and
        Operation Event Register set. It does not affect the Questionable Event
        Enable or Operation Event Enable registers.
        """
        self.write('*CLS')

    def reset(self) -> None:
        """
        Returns the instrument to default settings, cancels all pending
        commands.
        """
        self.write('*RST')
